/* BEGIN_COMMON_COPYRIGHT_HEADER
 * (c)LGPL2+
 *
 * Nde - a lightweight, Qt based, desktop toolset
 * https://lxqt.org
 *
 * Copyright: 2010-2011 Razor team
 * Authors:
 *   Alexander Sokoloff <sokoloff.a@gmail.com>
 *
 * This program or library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301 USA
 *
 * END_COMMON_COPYRIGHT_HEADER */


#include "ndepanel.h"
#include "ndepanellimits.h"
#include "indepanelplugin.h"
#include "ndepanelapplication.h"
#include "ndepanellayout.h"
#include "config/configpaneldialog.h"
#include "popupmenu.h"
#include "plugin.h"
#include "panelpluginsmodel.h"
#include "windownotifier.h"
#include "../plugin-taskbar/ndetaskbarplugin.h"
#include <Nde/PluginInfo>


#include <QScreen>
#include <QWindow>
#include <QX11Info>
#include <QDebug>
#include <QString>
#include <QDesktopWidget>
#include <QMenu>
#include <QMessageBox>
#include <QDropEvent>
#include <XdgIcon>
#include <XdgDirs>
#include <QProcess>
#include <KWindowSystem/KWindowSystem>
#include <KWindowSystem/NETWM>
#include <KWindowEffects>
// Turn on this to show the time required to load each plugin during startup
// #define DEBUG_PLUGIN_LOADTIME
#ifdef DEBUG_PLUGIN_LOADTIME
#include <QElapsedTimer>
#endif

// Config keys and groups
#define CFG_KEY_SCREENNUM          "desktop"
#define CFG_KEY_POSITION           "position"
#define CFG_KEY_PANELSIZE          "panelSize"
#define CFG_KEY_ICONSIZE           "iconSize"
#define CFG_KEY_LINECNT            "lineCount"
#define CFG_KEY_LENGTH             "width"
#define CFG_KEY_PERCENT            "width-percent"
#define CFG_KEY_ALIGNMENT          "alignment"
#define CFG_KEY_FONTCOLOR          "font-color"
#define CFG_KEY_BACKGROUNDCOLOR    "background-color"
#define CFG_KEY_BACKGROUNDIMAGE    "background-image"
#define CFG_KEY_OPACITY            "opacity"
#define CFG_KEY_RESERVESPACE       "reserve-space"
#define CFG_KEY_PLUGINS            "plugins"
#define CFG_KEY_HIDABLE            "hidable"
#define CFG_KEY_VISIBLE_MARGIN     "visible-margin"
#define CFG_KEY_ANIMATION          "animation-duration"
#define CFG_KEY_SHOW_DELAY         "show-delay"
#define CFG_KEY_LOCKPANEL          "lockPanel"

/************************************************
 Returns the Position by the string.
 String is one of "Top", "Left", "Bottom", "Right", string is not case sensitive.
 If the string is not correct, returns defaultValue.
 ************************************************/
INdePanel::Position NdePanel::strToPosition(const QString& str, INdePanel::Position defaultValue)
{
    if (str.toUpper() == "TOP")    return NdePanel::PositionTop;
    if (str.toUpper() == "LEFT")   return NdePanel::PositionLeft;
    if (str.toUpper() == "RIGHT")  return NdePanel::PositionRight;
    if (str.toUpper() == "BOTTOM") return NdePanel::PositionBottom;
    return defaultValue;
}

/************************************************
 Return  string representation of the position
 ************************************************/
QString NdePanel::positionToStr(INdePanel::Position position)
{
    switch (position)
    {
    case NdePanel::PositionTop:
        return QString("Top");
    case NdePanel::PositionLeft:
        return QString("Left");
    case NdePanel::PositionRight:
        return QString("Right");
    case NdePanel::PositionBottom:
        return QString("Bottom");
    }

    return QString();
}
QString NdePanel::positionToTr(INdePanel::Position position)
{
    switch (position)
    {
    case NdePanel::PositionTop:
        return QString(tr("Top"));
    case NdePanel::PositionLeft:
        return QString(tr("Left"));
    case NdePanel::PositionRight:
        return QString(tr("Right"));
    case NdePanel::PositionBottom:
        return QString(tr("Bottom"));
    }
    return QString();
}


/************************************************

 ************************************************/
NdePanel::NdePanel(const QString &configGroup, Nde::Settings *settings, QWidget *parent) :
    QFrame(parent),
    mSettings(settings),
    mConfigGroup(configGroup),
    mPlugins{nullptr},
    mStandaloneWindows{new WindowNotifier},
    mPanelSize(0),
    mIconSize(0),
    mLineCount(0),
    mLength(0),
    mAlignment(AlignmentLeft),
    mPosition(INdePanel::PositionBottom),
    mScreenNum(0), //whatever (avoid conditional on uninitialized value)
    mActualScreenNum(0),
    mHidable(false),
    mVisibleMargin(true),
    mHidden(false),
    mAnimationTime(0),
    mCompositorOpen(false),
    mReserveSpace(true),
    mAnimation(nullptr),
    mLockPanel(false)
{
    //You can find information about the flags and widget attributes in your
    //Qt documentation or at https://doc.qt.io/qt-5/qt.html
    //Qt::FramelessWindowHint = Produces a borderless window. The user cannot
    //move or resize a borderless window via the window system. On X11, ...
    //Qt::WindowStaysOnTopHint = Informs the window system that the window
    //should stay on top of all other windows. Note that on ...
    Qt::WindowFlags flags = Qt::FramelessWindowHint | Qt::WindowStaysOnTopHint;

    // NOTE: by PCMan:
    // In Qt 4, the window is not activated if it has Qt::WA_X11NetWmWindowTypeDock.
    // Since Qt 5, the default behaviour is changed. A window is always activated on mouse click.
    // Please see the source code of Qt5: src/plugins/platforms/xcb/qxcbwindow.cpp.
    // void QXcbWindow::handleButtonPressEvent(const xcb_button_press_event_t *event)
    // This new behaviour caused nde bug #161 - Cannot minimize windows from panel 1 when two task managers are open
    // Besides, this breaks minimizing or restoring windows when clicking on the taskbar buttons.
    // To workaround this regression bug, we need to add this window flag here.
    // However, since the panel gets no keyboard focus, this may decrease accessibility since
    // it's not possible to use the panel with keyboards. We need to find a better solution later.
    flags |= Qt::WindowDoesNotAcceptFocus;

    setWindowFlags(flags);
    //Adds _NET_WM_WINDOW_TYPE_DOCK to the window's _NET_WM_WINDOW_TYPE X11 window property. See https://standards.freedesktop.org/wm-spec/ for more details.
    setAttribute(Qt::WA_X11NetWmWindowTypeDock);
    //Enables tooltips for inactive windows.
    setAttribute(Qt::WA_AlwaysShowToolTips);
    //Indicates that the widget should have a translucent background, i.e., any non-opaque regions of the widgets will be translucent because the widget will have an alpha channel. Setting this ...
    setAttribute(Qt::WA_TranslucentBackground);
    //Allows data from drag and drop operations to be dropped onto the widget (see QWidget::setAcceptDrops()).
    setAttribute(Qt::WA_AcceptDrops);
    KWindowEffects::enableBlurBehind(winId(), true);

    setWindowTitle("Nde Panel");
    setObjectName(QString("NdePanel %1").arg(configGroup));

    //NdePanel (inherits QFrame) -> lav (QGridLayout) -> NdePanelWidget (QFrame) -> NdePanelLayout
    NdePanelWidget = new QFrame(this);
    NdePanelWidget->setObjectName("BackgroundWidget");
    QGridLayout* lav = new QGridLayout();
    lav->setContentsMargins(0, 0, 0, 0);
    setLayout(lav);
    this->layout()->addWidget(NdePanelWidget);

    mLayout = new NdePanelLayout(NdePanelWidget);
    connect(mLayout, &NdePanelLayout::pluginMoved, this, &NdePanel::pluginMoved);
    NdePanelWidget->setLayout(mLayout);
    mLayout->setLineCount(mLineCount);

    mDelaySave.setSingleShot(true);
    mDelaySave.setInterval(SETTINGS_SAVE_DELAY);
    connect(&mDelaySave, SIGNAL(timeout()), this, SLOT(saveSettings()));

    mHideTimer.setSingleShot(true);
    mHideTimer.setInterval(PANEL_HIDE_DELAY);
    connect(&mHideTimer, SIGNAL(timeout()), this, SLOT(hidePanelWork()));

    mShowDelayTimer.setSingleShot(true);
    mShowDelayTimer.setInterval(PANEL_SHOW_DELAY);
    connect(&mShowDelayTimer, &QTimer::timeout, [this] { showPanel(mAnimationTime > 0); });

    connect(QApplication::desktop(), &QDesktopWidget::resized, this, &NdePanel::ensureVisible);
    connect(QApplication::desktop(), &QDesktopWidget::screenCountChanged, this, &NdePanel::ensureVisible);

    // connecting to QDesktopWidget::workAreaResized shouldn't be necessary,
    // as we've already connceted to QDesktopWidget::resized, but it actually
    // is. Read mode on https://github.com/lxqt/nde-panel/pull/310
    connect(QApplication::desktop(), &QDesktopWidget::workAreaResized,
            this, &NdePanel::ensureVisible);

    connect(Nde::Settings::globalSettings(), SIGNAL(settingsChanged()), this, SLOT(update()));
    //颜色主题变化
//    connect(Nde::Settings::globalSettings(), &Nde::GlobalSettings::darkModeThemeChanged, [this](){
//        Q_EMIT realigned();
//        this->updateStyleSheet();
//    });
    connect(ndeApp, SIGNAL(themeChanged()), this, SLOT(realign()));
    connect(ndeApp, &QApplication::primaryScreenChanged, this, &NdePanel::onPrimaryScreenChange);

    connect(mStandaloneWindows.data(), &WindowNotifier::firstShown, [this] { showPanel(true); });
    connect(mStandaloneWindows.data(), &WindowNotifier::lastHidden, this, &NdePanel::hidePanel);

    QDBusInterface * compositer = new QDBusInterface(QString::fromLatin1("org.nde.session"),
                                                     QString::fromLatin1("/NdeSession"),
                                                     QString::fromLatin1("org.nde.session"),
                                                     QDBusConnection::sessionBus());
    if(compositer->isValid())
    {
        QDBusMessage reply = compositer->call("getCompositorOpen");
        if(!reply.arguments().isEmpty()) {
            mCompositorOpen =  reply.arguments().first().toBool();

        }
    }
    delete compositer;
    QDBusConnection::sessionBus().connect(QString::fromLatin1("org.nde.session"),
                                          QString::fromLatin1("/NdeSession"),
                                          QString::fromLatin1("org.nde.session"),
                                          QString::fromLatin1("compositorOpenStateChanged"),
                                          this, SLOT(onCompositorOpenChanged(bool)));
    NdePanelWidget->setProperty("compositor", mCompositorOpen);


    QDBusConnection connect = QDBusConnection::sessionBus();
    if(!connect.registerService("com.nde.panel.ndepanel"))
    {
        qWarning() << "D-BUS register service failed:com.nde.Taskbar";
    }
    if(!connect.registerObject("/com/nde/panel/ndepanel", this, QDBusConnection::ExportAllSlots))
    {
        qWarning() << "D-BUS register object failed:/org/nde/Taskbar";
    }

    readSettings();

    ensureVisible();

    loadPlugins();

    show();

    // show it the first time, despite setting
    if (mHidable)
    {
        showPanel(false);
        QTimer::singleShot(PANEL_HIDE_FIRST_TIME, this, SLOT(hidePanel()));
    }
}

/************************************************

 ************************************************/
void NdePanel::readSettings()
{
    // Read settings ......................................
    mSettings->beginGroup(mConfigGroup);

    // Let Hidability be the first thing we read
    // so that every call to realign() is without side-effect
    mHidable = mSettings->value(CFG_KEY_HIDABLE, mHidable).toBool();
    mHidden = mHidable;

    mVisibleMargin = mSettings->value(CFG_KEY_VISIBLE_MARGIN, mVisibleMargin).toBool();

    mAnimationTime = mSettings->value(CFG_KEY_ANIMATION, mAnimationTime).toInt();
    mShowDelayTimer.setInterval(mSettings->value(CFG_KEY_SHOW_DELAY, mShowDelayTimer.interval()).toInt());

    // By default we are using size & count from theme.
    setPanelSize(mSettings->value(CFG_KEY_PANELSIZE, PANEL_DEFAULT_SIZE).toInt(), false);
    setIconSize(mSettings->value(CFG_KEY_ICONSIZE, PANEL_DEFAULT_ICON_SIZE).toInt(), false);
    setLineCount(mSettings->value(CFG_KEY_LINECNT, PANEL_DEFAULT_LINE_COUNT).toInt(), false);

    setLength(mSettings->value(CFG_KEY_LENGTH, 100).toInt(),
              mSettings->value(CFG_KEY_PERCENT, true).toBool(),
              false);

    mScreenNum = mSettings->value(CFG_KEY_SCREENNUM, QApplication::desktop()->primaryScreen()).toInt();
//    mScreenNum = QApplication::desktop()->primaryScreen();
    setPosition(mScreenNum,
                strToPosition(mSettings->value(CFG_KEY_POSITION).toString(), PositionBottom),
                false);

    setAlignment(Alignment(mSettings->value(CFG_KEY_ALIGNMENT, mAlignment).toInt()), false);

    QColor color = mSettings->value(CFG_KEY_FONTCOLOR, "").value<QColor>();
    if (color.isValid())
        setFontColor(color, true);

    mReserveSpace = mSettings->value(CFG_KEY_RESERVESPACE, true).toBool();
    mLockPanel = mSettings->value(CFG_KEY_LOCKPANEL, false).toBool();

    mSettings->endGroup();
}


/************************************************

 ************************************************/
void NdePanel::saveSettings(bool later)
{
    mDelaySave.stop();
    if (later)
    {
        mDelaySave.start();
        return;
    }

    mSettings->beginGroup(mConfigGroup);

    //Note: save/load of plugin names is completely handled by mPlugins object
    //mSettings->setValue(CFG_KEY_PLUGINS, mPlugins->pluginNames());

    mSettings->setValue(CFG_KEY_PANELSIZE, mPanelSize);
    mSettings->setValue(CFG_KEY_ICONSIZE, mIconSize);
    mSettings->setValue(CFG_KEY_LINECNT, mLineCount);

    mSettings->setValue(CFG_KEY_LENGTH, mLength);
    mSettings->setValue(CFG_KEY_PERCENT, mLengthInPercents);

    mSettings->setValue(CFG_KEY_SCREENNUM, mScreenNum);
    mSettings->setValue(CFG_KEY_POSITION, positionToStr(mPosition));

    mSettings->setValue(CFG_KEY_ALIGNMENT, mAlignment);

    mSettings->setValue(CFG_KEY_FONTCOLOR, mFontColor.isValid() ? mFontColor : QColor());
    mSettings->setValue(CFG_KEY_RESERVESPACE, mReserveSpace);

    mSettings->setValue(CFG_KEY_HIDABLE, mHidable);
    mSettings->setValue(CFG_KEY_VISIBLE_MARGIN, mVisibleMargin);
    mSettings->setValue(CFG_KEY_ANIMATION, mAnimationTime);
    mSettings->setValue(CFG_KEY_SHOW_DELAY, mShowDelayTimer.interval());

    mSettings->setValue(CFG_KEY_LOCKPANEL, mLockPanel);

    mSettings->endGroup();
}


/************************************************

 ************************************************/
void NdePanel::ensureVisible()
{
    if (!canPlacedOn(mScreenNum, mPosition))
        setPosition(findAvailableScreen(mPosition), mPosition, false);
    else
        mActualScreenNum = mScreenNum;

    // the screen size might be changed
    realign();
}


/************************************************

 ************************************************/
NdePanel::~NdePanel()
{
    mLayout->setEnabled(false);
    delete mAnimation;
    delete mConfigDialog.data();
    // do not save settings because of "user deleted panel" functionality saveSettings();
}


/************************************************

 ************************************************/
void NdePanel::show()
{
    QWidget::show();
    KWindowSystem::setOnDesktop(effectiveWinId(), NET::OnAllDesktops);
}


/************************************************

 ************************************************/
QStringList pluginDesktopDirs()
{
    QStringList dirs;
    dirs << QString(getenv("NDE_PANEL_PLUGINS_DIR")).split(':', QString::SkipEmptyParts);
    dirs << QString("%1/%2").arg(XdgDirs::dataHome(), "/nde/nde-panel");
    dirs << PLUGIN_DESKTOPS_DIR;
    return dirs;
}


/************************************************

 ************************************************/
void NdePanel::loadPlugins()
{
    QString names_key(mConfigGroup);
    names_key += '/';
    names_key += QLatin1String(CFG_KEY_PLUGINS);
    mPlugins.reset(new PanelPluginsModel(this, names_key, pluginDesktopDirs()));

    connect(mPlugins.data(), &PanelPluginsModel::pluginAdded, mLayout, &NdePanelLayout::addPlugin);
    connect(mPlugins.data(), &PanelPluginsModel::pluginMovedUp, mLayout, &NdePanelLayout::moveUpPlugin);
    //reemit signals
    connect(mPlugins.data(), &PanelPluginsModel::pluginAdded, this, &NdePanel::pluginAdded);
    connect(mPlugins.data(), &PanelPluginsModel::pluginRemoved, this, &NdePanel::pluginRemoved);

    const auto plugins = mPlugins->plugins();
    for (auto const & plugin : plugins)
    {
        mLayout->addPlugin(plugin);
        connect(plugin, &Plugin::dragLeft, [this] { mShowDelayTimer.stop(); hidePanel(); });
    }
}

/************************************************

 ************************************************/
int NdePanel::getReserveDimension()
{
    return mHidable ? PANEL_HIDE_SIZE : qMax(PANEL_MINIMUM_SIZE, mPanelSize);
}

void NdePanel::setPanelGeometry(bool animate)
{
    const QRect currentScreen = QApplication::desktop()->screenGeometry(mActualScreenNum);
    QRect rect;

    if (isHorizontal())
    {
        // Horiz panel ***************************
        rect.setHeight(qMax(PANEL_MINIMUM_SIZE, mPanelSize));
        if (mLengthInPercents)
            rect.setWidth(currentScreen.width() * mLength / 100.0);
        else
        {
            if (mLength <= 0)
                rect.setWidth(currentScreen.width() + mLength);
            else
                rect.setWidth(mLength);
        }

        rect.setWidth(qMax(rect.size().width(), mLayout->minimumSize().width()));

        // Horiz ......................
        switch (mAlignment)
        {
        case NdePanel::AlignmentLeft:
            rect.moveLeft(currentScreen.left());
            break;

        case NdePanel::AlignmentCenter:
            rect.moveCenter(currentScreen.center());
            break;

        case NdePanel::AlignmentRight:
            rect.moveRight(currentScreen.right());
            break;
        }

        // Vert .......................
        if (mPosition == INdePanel::PositionTop)
        {
            if (mHidden)
                rect.moveBottom(currentScreen.top() + PANEL_HIDE_SIZE - 1);
            else
                rect.moveTop(currentScreen.top());
        }
        else
        {
            if (mHidden)
                rect.moveTop(currentScreen.bottom() - PANEL_HIDE_SIZE + 1);
            else
                rect.moveBottom(currentScreen.bottom());
        }
    }
    else
    {
        // Vert panel ***************************
        rect.setWidth(qMax(PANEL_MINIMUM_SIZE, mPanelSize));
        if (mLengthInPercents)
            rect.setHeight(currentScreen.height() * mLength / 100.0);
        else
        {
            if (mLength <= 0)
                rect.setHeight(currentScreen.height() + mLength);
            else
                rect.setHeight(mLength);
        }

        rect.setHeight(qMax(rect.size().height(), mLayout->minimumSize().height()));

        // Vert .......................
        switch (mAlignment)
        {
        case NdePanel::AlignmentLeft:
            rect.moveTop(currentScreen.top());
            break;

        case NdePanel::AlignmentCenter:
            rect.moveCenter(currentScreen.center());
            break;

        case NdePanel::AlignmentRight:
            rect.moveBottom(currentScreen.bottom());
            break;
        }

        // Horiz ......................
        if (mPosition == INdePanel::PositionLeft)
        {
            if (mHidden)
                rect.moveRight(currentScreen.left() + PANEL_HIDE_SIZE - 1);
            else
                rect.moveLeft(currentScreen.left());
        }
        else
        {
            if (mHidden)
                rect.moveLeft(currentScreen.right() - PANEL_HIDE_SIZE + 1);
            else
                rect.moveRight(currentScreen.right());
        }
    }
    if (rect != geometry())
    {
        setFixedSize(rect.size());
        if (animate)
        {
            if (mAnimation == nullptr)
            {
                mAnimation = new QPropertyAnimation(this, "geometry");
                mAnimation->setEasingCurve(QEasingCurve::Linear);
                //Note: for hiding, the margins are set after animation is finished
                connect(mAnimation, &QAbstractAnimation::finished, [this] { if (mHidden) setMargins(); });
            }
            mAnimation->setDuration(mAnimationTime);
            mAnimation->setStartValue(geometry());
            mAnimation->setEndValue(rect);
            //Note: for showing-up, the margins are removed instantly
            if (!mHidden)
                setMargins();
            mAnimation->start();
        }
        else
        {
            setMargins();
            setGeometry(rect);
	    //setFixedSize及SetGeometry之后LayoutRequest事件在Centos8上无法收到，导致布局无法更新
	    QApplication::postEvent(this, new QEvent(QEvent::LayoutRequest));
        }
    }
}

void NdePanel::setMargins()
{
    if (mHidden)
    {
        if (isHorizontal())
        {
            if (mPosition == INdePanel::PositionTop)
                mLayout->setContentsMargins(0, 0, 0, PANEL_HIDE_SIZE);
            else
                mLayout->setContentsMargins(0, PANEL_HIDE_SIZE, 0, 0);
        }
        else
        {
            if (mPosition == INdePanel::PositionLeft)
                mLayout->setContentsMargins(0, 0, PANEL_HIDE_SIZE, 0);
            else
                mLayout->setContentsMargins(PANEL_HIDE_SIZE, 0, 0, 0);
        }
        if (!mVisibleMargin)
            setWindowOpacity(0.0);
    }
    else {
        mLayout->setContentsMargins(0, 0, 0, 0);
        if (!mVisibleMargin)
            setWindowOpacity(1.0);
    }
}

void NdePanel::realign()
{
    if (!isVisible())
        return;
#if 0
    qDebug() << "** Realign *********************";
    qDebug() << "PanelSize:   " << mPanelSize;
    qDebug() << "IconSize:      " << mIconSize;
    qDebug() << "LineCount:     " << mLineCount;
    qDebug() << "Length:        " << mLength << (mLengthInPercents ? "%" : "px");
    qDebug() << "Alignment:     " << (mAlignment == 0 ? "center" : (mAlignment < 0 ? "left" : "right"));
    qDebug() << "Position:      " << positionToStr(mPosition) << "on" << mScreenNum;
//    qDebug() << "Plugins count: " << mPlugins.count();
#endif

    setPanelGeometry();

    // Reserve our space on the screen ..........
    // It's possible that our geometry is not changed, but screen resolution is changed,
    // so resetting WM_STRUT is still needed. To make it simple, we always do it.
    updateWmStrut();
}


// Update the _NET_WM_PARTIAL_STRUT and _NET_WM_STRUT properties for the window
void NdePanel::updateWmStrut()
{
    WId wid = effectiveWinId();
    if(wid == 0 || !isVisible())
        return;

    if (mReserveSpace)
    {
        const QRect wholeScreen = QApplication::desktop()->geometry();
        const QRect rect = geometry();
        // NOTE: https://standards.freedesktop.org/wm-spec/wm-spec-latest.html
        // Quote from the EWMH spec: " Note that the strut is relative to the screen edge, and not the edge of the xinerama monitor."
        // So, we use the geometry of the whole screen to calculate the strut rather than using the geometry of individual monitors.
        // Though the spec only mention Xinerama and did not mention XRandR, the rule should still be applied.
        // At least openbox is implemented like this.
        switch (mPosition)
        {
        case NdePanel::PositionTop:
            KWindowSystem::setExtendedStrut(wid,
                                            /* Left   */  0, 0, 0,
                                            /* Right  */  0, 0, 0,
                                            /* Top    */  rect.top() + getReserveDimension(), rect.left(), rect.right(),
                                            /* Bottom */  0, 0, 0
                                           );
            break;

        case NdePanel::PositionBottom:
            KWindowSystem::setExtendedStrut(wid,
                                            /* Left   */  0, 0, 0,
                                            /* Right  */  0, 0, 0,
                                            /* Top    */  0, 0, 0,
                                            /* Bottom */  wholeScreen.bottom() - rect.bottom() + getReserveDimension(), rect.left(), rect.right()
                                           );
            break;

        case NdePanel::PositionLeft:
            KWindowSystem::setExtendedStrut(wid,
                                            /* Left   */  rect.left() + getReserveDimension(), rect.top(), rect.bottom(),
                                            /* Right  */  0, 0, 0,
                                            /* Top    */  0, 0, 0,
                                            /* Bottom */  0, 0, 0
                                           );

            break;

        case NdePanel::PositionRight:
            KWindowSystem::setExtendedStrut(wid,
                                            /* Left   */  0, 0, 0,
                                            /* Right  */  wholeScreen.right() - rect.right() + getReserveDimension(), rect.top(), rect.bottom(),
                                            /* Top    */  0, 0, 0,
                                            /* Bottom */  0, 0, 0
                                           );
            break;
    }
    } else
    {
        KWindowSystem::setExtendedStrut(wid,
                                        /* Left   */  0, 0, 0,
                                        /* Right  */  0, 0, 0,
                                        /* Top    */  0, 0, 0,
                                        /* Bottom */  0, 0, 0
                                       );
    }
}


/************************************************
  The panel can't be placed on boundary of two displays.
  This function checks if the panel can be placed on the display
  @screenNum on @position.
 ************************************************/
bool NdePanel::canPlacedOn(int screenNum, NdePanel::Position position)
{
    QDesktopWidget* dw = QApplication::desktop();

    switch (position)
    {
    case NdePanel::PositionTop:
        for (int i = 0; i < dw->screenCount(); ++i)
            if (dw->screenGeometry(i).bottom() < dw->screenGeometry(screenNum).top())
                return false;
        return true;

    case NdePanel::PositionBottom:
        for (int i = 0; i < dw->screenCount(); ++i)
            if (dw->screenGeometry(i).top() > dw->screenGeometry(screenNum).bottom())
                return false;
        return true;

    case NdePanel::PositionLeft:
        for (int i = 0; i < dw->screenCount(); ++i)
            if (dw->screenGeometry(i).right() < dw->screenGeometry(screenNum).left())
                return false;
        return true;

    case NdePanel::PositionRight:
        for (int i = 0; i < dw->screenCount(); ++i)
            if (dw->screenGeometry(i).left() > dw->screenGeometry(screenNum).right())
                return false;
        return true;
    }

    return false;
}


/************************************************

 ************************************************/
int NdePanel::findAvailableScreen(NdePanel::Position position)
{
    int current = mScreenNum;

    for (int i = current; i < QApplication::desktop()->screenCount(); ++i)
        if (canPlacedOn(i, position))
            return i;

    for (int i = 0; i < current; ++i)
        if (canPlacedOn(i, position))
            return i;

    return 0;
}


/************************************************

 ************************************************/
void NdePanel::showConfigDialog()
{
    if (mConfigDialog.isNull())
        mConfigDialog = new ConfigPanelDialog(this, nullptr /*make it top level window*/);

    mConfigDialog->showConfigPanelPage();
    mStandaloneWindows->observeWindow(mConfigDialog.data());
    mConfigDialog->show();
    mConfigDialog->raise();
    mConfigDialog->activateWindow();
    WId wid = mConfigDialog->windowHandle()->winId();

    KWindowSystem::activateWindow(wid);
    KWindowSystem::setOnDesktop(wid, KWindowSystem::currentDesktop());
}


/************************************************

 ************************************************/
void NdePanel::showAddPluginDialog()
{
    if (mConfigDialog.isNull())
        mConfigDialog = new ConfigPanelDialog(this, nullptr /*make it top level window*/);

    mConfigDialog->showConfigPluginsPage();
    mStandaloneWindows->observeWindow(mConfigDialog.data());
    mConfigDialog->show();
    mConfigDialog->raise();
    mConfigDialog->activateWindow();
    WId wid = mConfigDialog->windowHandle()->winId();

    KWindowSystem::activateWindow(wid);
    KWindowSystem::setOnDesktop(wid, KWindowSystem::currentDesktop());
}
void NdePanel::showTaskManager()
{
    QProcess::startDetached("stacer");
}

void NdePanel::setPluginVisible(QString pluginName)
{
    Plugin * plugin = mPlugins->pluginByName(pluginName);
    if(plugin)
    {
        INdePanelPlugin * iPlugin = const_cast<INdePanelPlugin *>( plugin->iPlugin());
        if(iPlugin)
        {
            iPlugin->setPluginVisible(!iPlugin->getPluginVisible());
            QDBusInterface* interface = new QDBusInterface(QString::fromLatin1("org.nde.systemsettings.Taskbar"),
                                                             QString::fromLatin1("/org/nde/Taskbar"),
                                                             QString::fromLatin1("org.nde.systemsettings.Taskbar"),
                                                             QDBusConnection::sessionBus());
            interface->call("onPluginModeChange", pluginName);
            delete interface;
        }
    }
}

/************************************************

 ************************************************/
void NdePanel::updateStyleSheet()
{
    //
}

/************************************************

 ************************************************/
void NdePanel::setPanelSize(int value, bool save)
{
    if (mPanelSize != value)
    {
        mPanelSize = value;
        realign();

        if (save)
            saveSettings(true);
    }
}

/************************************************

 ************************************************/
void NdePanel::setIconSize(int value, bool save)
{
    if (mIconSize != value)
    {
        mIconSize = value;
        updateStyleSheet();
        mLayout->setLineSize(mIconSize);

        if (save)
            saveSettings(true);

        realign();
    }
}


/************************************************

 ************************************************/
void NdePanel::setLineCount(int value, bool save)
{
    if (mLineCount != value)
    {
        mLineCount = value;
        mLayout->setEnabled(false);
        mLayout->setLineCount(mLineCount);
        mLayout->setEnabled(true);

        if (save)
            saveSettings(true);

        realign();
    }
}


/************************************************

 ************************************************/
void NdePanel::setLength(int length, bool inPercents, bool save)
{
    if (mLength == length &&
            mLengthInPercents == inPercents)
        return;

    mLength = length;
    mLengthInPercents = inPercents;

    if (save)
        saveSettings(true);

    realign();
}


/************************************************

 ************************************************/
void NdePanel::setPosition(int screen, INdePanel::Position position, bool save)
{
    if (mScreenNum == screen &&
            mPosition == position)
        return;

    mActualScreenNum = screen;
    mPosition = position;
    mLayout->setPosition(mPosition);

    if (save)
    {
        mScreenNum = screen;
        saveSettings(true);
    }

    // Qt 5 adds a new class QScreen and add API for setting the screen of a QWindow.
    // so we had better use it. However, without this, our program should still work
    // as long as XRandR is used. Since XRandR combined all screens into a large virtual desktop
    // every screen and their virtual siblings are actually on the same virtual desktop.
    // So things still work if we don't set the screen correctly, but this is not the case
    // for other backends, such as the upcoming wayland support. Hence it's better to set it.
    if(windowHandle())
    {
        // QScreen* newScreen = qApp->screens().at(screen);
        // QScreen* oldScreen = windowHandle()->screen();
        // const bool shouldRecreate = windowHandle()->handle() && !(oldScreen && oldScreen->virtualSiblings().contains(newScreen));
        // Q_ASSERT(shouldRecreate == false);

        // NOTE: When you move a window to another screen, Qt 5 might recreate the window as needed
        // But luckily, this never happen in XRandR, so Qt bug #40681 is not triggered here.
        // (The only exception is when the old screen is destroyed, Qt always re-create the window and
        // this corner case triggers #40681.)
        // When using other kind of multihead settings, such as Xinerama, this might be different and
        // unless Qt developers can fix their bug, we have no way to workaround that.
        windowHandle()->setScreen(qApp->screens().at(screen));
    }

    realign();
}

void NdePanel::onPositionSetting(int index)
{
    setPosition(mScreenNum, (INdePanel::Position)index, true);
}

/************************************************
 *
 ************************************************/
void NdePanel::setAlignment(Alignment value, bool save)
{
    if (mAlignment == value)
        return;

    mAlignment = value;

    if (save)
        saveSettings(true);

    realign();
}

/************************************************
 *
 ************************************************/
void NdePanel::setFontColor(QColor color, bool save)
{
    mFontColor = color;
    updateStyleSheet();

    if (save)
        saveSettings(true);
}

/************************************************
 *
 ************************************************/
void NdePanel::setReserveSpace(bool reserveSpace, bool save)
{
    if (mReserveSpace == reserveSpace)
        return;

    mReserveSpace = reserveSpace;

    if (save)
        saveSettings(true);

    updateWmStrut();
}


/************************************************

 ************************************************/
QRect NdePanel::globalGeometry() const
{
    // panel is the the top-most widget/window, no calculation needed
    return geometry();
}


/************************************************

 ************************************************/
bool NdePanel::event(QEvent *event)
{
    switch (event->type())
    {
    case QEvent::ContextMenu:
        showPopupMenu();
        break;

    case QEvent::LayoutRequest:
        Q_EMIT realigned();
        break;

    case QEvent::WinIdChange:
    {
        // qDebug() << "WinIdChange" << hex << effectiveWinId();
        if(effectiveWinId() == 0)
            break;

        // Sometimes Qt needs to re-create the underlying window of the widget and
        // the winId() may be changed at runtime. So we need to reset all X11 properties
        // when this happens.
        qDebug() << "WinIdChange" << hex << effectiveWinId() << "handle" << windowHandle() << windowHandle()->screen();

        // Qt::WA_X11NetWmWindowTypeDock becomes ineffective in Qt 5
        // See QTBUG-39887: https://bugreports.qt-project.org/browse/QTBUG-39887
        // Let's use KWindowSystem for that
        KWindowSystem::setType(effectiveWinId(), NET::Dock);

        updateWmStrut(); // reserve screen space for the panel
        KWindowSystem::setOnAllDesktops(effectiveWinId(), true);
        break;
    }
    case QEvent::DragEnter:
        dynamic_cast<QDropEvent *>(event)->setDropAction(Qt::IgnoreAction);
        event->accept();
        //no break intentionally
    case QEvent::Enter:
        mShowDelayTimer.start();
        break;

    case QEvent::Leave:
    case QEvent::DragLeave:
        mShowDelayTimer.stop();
        hidePanel();
        break;

    default:
        break;
    }

    return QFrame::event(event);
}

/************************************************

 ************************************************/

void NdePanel::showEvent(QShowEvent *event)
{
    QFrame::showEvent(event);
    realign();
}


/************************************************

 ************************************************/
void NdePanel::showPopupMenu(Plugin *plugin)
{
//    Q_UNUSED(plugin)
    PopupMenu * menu = new PopupMenu(compositorOpen(),  this);
    menu->setAttribute(Qt::WA_DeleteOnClose);

    //任务管理器
    menu->addAction(tr("Task Manager"),
                    this, SLOT(showTaskManager()));

    menu->addSeparator();

    //桌面切换
    bool visibleDesktopSwitch = true;
    Plugin * pluginDesktopSwitch = mPlugins->pluginByName("desktopswitch");
    if(pluginDesktopSwitch) visibleDesktopSwitch = pluginDesktopSwitch->iPlugin()->getPluginVisible();
    QAction *dsAction = menu->addAction(tr("Show \"Multi Views\" Button"),
                    [this](){setPluginVisible("desktopswitch");});
    dsAction->setCheckable(true);
    dsAction->setChecked(visibleDesktopSwitch ? true:false);

    //快速启动栏
    bool visibleQuickLaunch = true;
    Plugin * pluginQuickLaunch = mPlugins->pluginByName("quicklaunch");
    if(pluginQuickLaunch) visibleQuickLaunch = pluginQuickLaunch->iPlugin()->getPluginVisible();
    QAction *qlAction = menu->addAction(tr("Show \"Quick Launch\""),
                    [this](){setPluginVisible("quicklaunch");});
    qlAction->setCheckable(true);
    qlAction->setChecked(visibleQuickLaunch ? true:false);

    menu->addSeparator();

    //设置平铺模式
    bool isIconMode = true;
    Plugin * pluginTaskBar = mPlugins->pluginByName("taskbar");
    if(pluginTaskBar)
    {
        INdePanelPlugin * iPlugin = const_cast<INdePanelPlugin *>( pluginTaskBar->iPlugin());
        NdeTaskBarPlugin * ndeTaskBarPlugin = static_cast<NdeTaskBarPlugin *>(iPlugin);
        if(ndeTaskBarPlugin)
        {
            isIconMode = ndeTaskBarPlugin->isIconMode();
        }
    }
    QAction * taskBarAction = menu->addAction(tr("TileMode"),
                                              [this](bool checked){
        Plugin * pluginTaskBar = mPlugins->pluginByName("taskbar");
        if(pluginTaskBar)
        {
            INdePanelPlugin * iPlugin = const_cast<INdePanelPlugin *>( pluginTaskBar->iPlugin());
            NdeTaskBarPlugin * ndeTaskBarPlugin = static_cast<NdeTaskBarPlugin *>(iPlugin);
            if(ndeTaskBarPlugin)
            {
                ndeTaskBarPlugin->setIconMode(!checked);
                // 当模式改变时发送信号给控制中心的任务栏模块
                QDBusInterface* interface = new QDBusInterface(QString::fromLatin1("org.nde.systemsettings.Taskbar"),
                                    QString::fromLatin1("/org/nde/Taskbar"),
                                    QString::fromLatin1("org.nde.systemsettings.Taskbar"),
                                    QDBusConnection::sessionBus());
                interface->call("onTaskbarButton", !checked ? "Icon": "IconText");
                delete interface;
            }
        }
    });
    taskBarAction->setCheckable(true);
    taskBarAction->setChecked(!isIconMode);

    //自动隐藏panel
    QAction *hAction = menu->addAction(tr("A&uto-hide"));
    hAction->setCheckable(true);
    if(mHidable)
    {
        hAction->setChecked(true);
    }
    connect(hAction, &QAction::triggered, this, &NdePanel::onAutoHideTriggered);

    //panel位置设置
    PopupMenu * layout_menu = new PopupMenu(compositorOpen(),  this);
    layout_menu->setTitle(tr("Task Layout"));
    menu->addMenu(layout_menu);
    for (int i =0 ; i <= NdePanel::PositionRight ; i++) {
        NdePanel::Position pos = static_cast<NdePanel::Position>(i);
        QAction * action = layout_menu->addAction(positionToTr(pos),
                                                  this, [=]{
            setPosition(mScreenNum,
                        pos, true);
            // 当位置改变时发送信号给控制中心的任务栏模块
            QDBusInterface* interface = new QDBusInterface(QString::fromLatin1("org.nde.systemsettings.Taskbar"),
                                    QString::fromLatin1("/org/nde/Taskbar"),
                                    QString::fromLatin1("org.nde.systemsettings.Taskbar"),
                                    QDBusConnection::sessionBus());
            interface->call("onPositionPanel", (int)pos);
            delete interface;
        });
        if(mPosition == i)
        {
            action->setCheckable(true);
            action->setChecked(true);
        }
    }

#ifdef DEBUG
    menu->addSeparator();
    menu->addAction("Exit (debug only)", qApp, SLOT(quit()));
#endif

    /* Note: in multihead & multipanel setup the QMenu::popup/exec places the window
     * sometimes wrongly (it seems that this bug is somehow connected to misinterpretation
     * of QDesktopWidget::availableGeometry)
     */
    menu->setGeometry(calculatePopupWindowPos(QCursor::pos(), menu->sizeHint()));
    willShowWindow(menu);
    menu->show();
}

void NdePanel::onAutoHideTriggered(bool checked)
{
    QDBusInterface* interface = new QDBusInterface(QString::fromLatin1("org.nde.systemsettings.Taskbar"),
                                                 QString::fromLatin1("/org/nde/Taskbar"),
                                                 QString::fromLatin1("org.nde.systemsettings.Taskbar"),
                                                 QDBusConnection::sessionBus());
    interface->call("onToggleHideDesktopMode", checked);
    delete interface;

    setHidable(checked, true);
    setReserveSpace(!checked, true);
    checked ? mHideTimer.start() : mHideTimer.stop();
}

void NdePanel::onAutoHideSetting(bool checked)
{
    if(!checked) // 如果取消隐藏任务栏，需先显示任务栏
    {
        showPanel(true);
    }
    setHidable(checked, true);
    setReserveSpace(!checked, true);
    checked ? mHideTimer.start() : mHideTimer.stop();
}

void NdePanel::onpluginTaskBar(int index)
{
    Plugin * pluginTaskBar = mPlugins->pluginByName("taskbar");
    if(pluginTaskBar)
    {
        INdePanelPlugin * iPlugin = const_cast<INdePanelPlugin *>( pluginTaskBar->iPlugin());
        NdeTaskBarPlugin * ndeTaskBarPlugin = static_cast<NdeTaskBarPlugin *>(iPlugin);
        if(ndeTaskBarPlugin)
        {
            ndeTaskBarPlugin->setIconMode(!index);
        }
    }
}

Plugin* NdePanel::findPlugin(const INdePanelPlugin* iPlugin) const
{
    const auto plugins = mPlugins->plugins();
    for (auto const & plug : plugins)
        if (plug->iPlugin() == iPlugin)
            return plug;
    return nullptr;
}

const INdePanelPlugin *NdePanel::getPlugin(const QString id)
{
    const auto plugins = mPlugins->plugins();
    for (auto const & plug : plugins)
    {
        if (plug->iPlugin()->themeId() == id)
            return plug->iPlugin();
    }
    return nullptr;
}

/************************************************

 ************************************************/
QRect NdePanel::calculatePopupWindowPos(QPoint const & absolutePos, QSize const & windowSize) const
{
    int x = absolutePos.x(), y = absolutePos.y();
    QRect screen = QApplication::desktop()->screenGeometry(this);

    switch (position())
    {
    case INdePanel::PositionTop:
        y = globalGeometry().bottom();
        screen.adjust(0,this->panelSize(),0,0);
        break;

    case INdePanel::PositionBottom:
        y = globalGeometry().top() - windowSize.height();
        screen.adjust(0,0,0,-this->panelSize());
        break;

    case INdePanel::PositionLeft:
        x = globalGeometry().right();
        screen.adjust(this->panelSize(),0,0,0);
        break;

    case INdePanel::PositionRight:
        x = globalGeometry().left() - windowSize.width();
        screen.adjust(0,0,-this->panelSize(),0);
        break;
    }
    QRect res(QPoint(x, y), windowSize);

    // NOTE: We cannot use AvailableGeometry() which returns the work area here because when in a
    // multihead setup with different resolutions. In this case, the size of the work area is limited
    // by the smallest monitor and may be much smaller than the current screen and we will place the
    // menu at the wrong place. This is very bad for UX. So let's use the full size of the screen.
    if (res.right() > screen.right())
        res.moveRight(screen.right());

    if (res.bottom() > screen.bottom())
        res.moveBottom(screen.bottom());

    if (res.left() < screen.left())
        res.moveLeft(screen.left());

    if (res.top() < screen.top())
        res.moveTop(screen.top());

    return res;
}

/************************************************

 ************************************************/
QRect NdePanel::calculatePopupWindowPos(const INdePanelPlugin *plugin, const QSize &windowSize) const
{
    Plugin *panel_plugin = findPlugin(plugin);
    if (nullptr == panel_plugin)
    {
        qWarning() << Q_FUNC_INFO << "Wrong logic? Unable to find Plugin* for" << plugin << "known plugins follow...";
        const auto plugins = mPlugins->plugins();
        for (auto const & plug : plugins)
            qWarning() << plug->iPlugin() << plug;

        return QRect();
    }

    // Note: assuming there are not contentMargins around the "BackgroundWidget" (NdePanelWidget)
    return calculatePopupWindowPos(globalGeometry().topLeft() + panel_plugin->geometry().topLeft(), windowSize);
}


/************************************************

 ************************************************/
void NdePanel::willShowWindow(QWidget * w)
{
    mStandaloneWindows->observeWindow(w);
}

/************************************************

 ************************************************/
void NdePanel::pluginFlagsChanged(const INdePanelPlugin * /*plugin*/)
{
    mLayout->rebuild();
}

/************************************************

 ************************************************/
QString NdePanel::qssPosition() const
{
    return positionToStr(position());
}

/************************************************

 ************************************************/
void NdePanel::pluginMoved(Plugin * plug)
{
    //get new position of the moved plugin
    bool found{false};
    QString plug_is_before;
    for (int i=0; i<mLayout->count(); ++i)
    {
        Plugin *plugin = qobject_cast<Plugin*>(mLayout->itemAt(i)->widget());
        if (plugin)
        {
            if (found)
            {
                //we found our plugin in previous cycle -> is before this (or empty as last)
                plug_is_before = plugin->settingsGroup();
                break;
            } else
                found = (plug == plugin);
        }
    }
    mPlugins->movePlugin(plug, plug_is_before);
}


/************************************************

 ************************************************/
void NdePanel::userRequestForDeletion()
{
    const QMessageBox::StandardButton ret
        = QMessageBox::warning(this, tr("Remove Panel", "Dialog Title") ,
            tr("Removing a panel can not be undone.\nDo you want to remove this panel?"),
            QMessageBox::Yes | QMessageBox::No);

    if (ret != QMessageBox::Yes) {
        return;
    }

    mSettings->beginGroup(mConfigGroup);
    const QStringList plugins = mSettings->value("plugins").toStringList();
    mSettings->endGroup();

    for(const QString& i : plugins)
        if (!i.isEmpty())
            mSettings->remove(i);

    mSettings->remove(mConfigGroup);

    Q_EMIT deletedByUser(this);
}

void NdePanel::onCompositorOpenChanged(bool open)
{
    if(mCompositorOpen == open) return;

    mCompositorOpen = open;
    NdePanelWidget->setProperty("compositor", open);
    NdePanelWidget->style()->unpolish(NdePanelWidget);
    NdePanelWidget->style()->polish(NdePanelWidget);
    Q_EMIT realigned();
}

void NdePanel::onPrimaryScreenChange(QScreen *screen)
{
    //
    mScreenNum = QApplication::desktop()->primaryScreen();
    setPosition(mScreenNum,
                mPosition,
                false);

    ensureVisible();
}

void NdePanel::showPanel(bool animate)
{
    if (mHidable)
    {
        mHideTimer.stop();
        if (mHidden)
        {
            mHidden = false;
            setPanelGeometry(mAnimationTime > 0 && animate);
        }
    }
}

void NdePanel::hidePanel()
{
    if (mHidable && !mHidden
            && !mStandaloneWindows->isAnyWindowShown()
       )
        mHideTimer.start();
}

void NdePanel::hidePanelWork()
{
    if (!geometry().contains(QCursor::pos()))
    {
        if (!mStandaloneWindows->isAnyWindowShown())
        {
            mHidden = true;
            setPanelGeometry(mAnimationTime > 0);
        } else
        {
            mHideTimer.start();
        }
    }
}

void NdePanel::setHidable(bool hidable, bool save)
{
    if (mHidable == hidable)
        return;

    mHidable = hidable;

    if (save)
        saveSettings(true);

    realign();
}

void NdePanel::setVisibleMargin(bool visibleMargin, bool save)
{
    if (mVisibleMargin == visibleMargin)
        return;

    mVisibleMargin = visibleMargin;

    if (save)
        saveSettings(true);

    realign();
}

void NdePanel::setAnimationTime(int animationTime, bool save)
{
    if (mAnimationTime == animationTime)
        return;

    mAnimationTime = animationTime;

    if (save)
        saveSettings(true);
}

void NdePanel::setShowDelay(int showDelay, bool save)
{
    if (mShowDelayTimer.interval() == showDelay)
        return;

    mShowDelayTimer.setInterval(showDelay);

    if (save)
        saveSettings(true);
}

QString NdePanel::iconTheme() const
{
    return mSettings->value("iconTheme").toString();
}

void NdePanel::setIconTheme(const QString& iconTheme)
{
    NdePanelApplication *a = reinterpret_cast<NdePanelApplication*>(qApp);
    a->setIconTheme(iconTheme);
}

void NdePanel::updateConfigDialog() const
{
    if (!mConfigDialog.isNull() && mConfigDialog->isVisible())
    {
        mConfigDialog->updateIconThemeSettings();
        const QList<QWidget*> widgets = mConfigDialog->findChildren<QWidget*>();
        for (QWidget *widget : widgets)
            widget->update();
    }
}

bool NdePanel::isPluginSingletonAndRunnig(QString const & pluginId) const
{
    Plugin const * plugin = mPlugins->pluginByID(pluginId);
    if (nullptr == plugin)
        return false;
    else
        return plugin->iPlugin()->flags().testFlag(INdePanelPlugin::SingleInstance);
}

void NdePanel::onpluginVisible(QString pluginName)
{
    Plugin * plugin = mPlugins->pluginByName(pluginName);
    if(plugin)
    {
        INdePanelPlugin * iPlugin = const_cast<INdePanelPlugin *>( plugin->iPlugin());
        if(iPlugin)
        {
            iPlugin->setPluginVisible(!iPlugin->getPluginVisible());
        }
    }
}

void NdePanel::setTaskbarSizeBySystemsetting(QString type)
{
    if(type == "Small")
    {
        setPanelSize(PANEL_DEFAULT_SIZE, false);
        setIconSize(PANEL_DEFAULT_ICON_SIZE, true);
        Plugin * pluginTaskBar = mPlugins->pluginByName("taskbar");
        if(pluginTaskBar)
        {
            INdePanelPlugin * iPlugin = const_cast<INdePanelPlugin *>( pluginTaskBar->iPlugin());
            NdeTaskBarPlugin * ndeTaskBarPlugin = static_cast<NdeTaskBarPlugin *>(iPlugin);
            if(ndeTaskBarPlugin)
            {
                ndeTaskBarPlugin->panelSizeChanged();
            }
        }
    }
    else if(type == "Big")
    {
        setPanelSize(PANEL_BIG_SIZE, false);
        setIconSize(PANEL_BIG_ICON_SIZE, true);
        Plugin * pluginTaskBar = mPlugins->pluginByName("taskbar");
        if(pluginTaskBar)
        {
            INdePanelPlugin * iPlugin = const_cast<INdePanelPlugin *>( pluginTaskBar->iPlugin());
            NdeTaskBarPlugin * ndeTaskBarPlugin = static_cast<NdeTaskBarPlugin *>(iPlugin);
            if(ndeTaskBarPlugin)
            {
                ndeTaskBarPlugin->panelSizeChanged();
            }
        }
    }
}
