/* BEGIN_COMMON_COPYRIGHT_HEADER
 * (c)LGPL2+
 *
 * Nde - a lightweight, Qt based, desktop toolset
 * https://lxqt.org
 *
 * Copyright: 2010-2011 Razor team
 * Authors:
 *   Alexander Sokoloff <sokoloff.a@gmail.com>
 *
 * This program or library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301 USA
 *
 * END_COMMON_COPYRIGHT_HEADER */


#ifndef NDEPANELAPPLICATION_H
#define NDEPANELAPPLICATION_H

#include <Nde/Application>
#include "indepanelplugin.h"

class QScreen;

class NdePanel;
class NdePanelApplicationPrivate;

/*!
 * \brief The NdePanelApplication class inherits from Nde::Application and
 * is therefore the QApplication that we will create and execute in our
 * main()-function.
 *
 * NdePanelApplication itself is not a visible panel, rather it is only
 * the container which holds the visible panels. These visible panels are
 * NdePanel objects which are stored in mPanels. This approach enables us
 * to have more than one panel (for example one panel at the top and one
 * panel at the bottom of the screen) without additional effort.
 */
class NdePanelApplication : public Nde::Application
{
    Q_OBJECT
    Q_CLASSINFO("D-Bus Interface", "com.nde.panel.ndepanelapplication")
public:
    /*!
     * \brief Creates a new NdePanelApplication with the given command line
     * arguments. Performs the following steps:
     * 1. Initializes the Nde::Application, sets application name and version.
     * 2. Handles command line arguments. Currently, the only cmdline argument
     * is -c = -config = -configfile which chooses a different config file
     * for the Nde::Settings.
     * 3. Creates the Nde::Settings.
     * 4. Connects QCoreApplication::aboutToQuit to cleanup().
     * 5. Calls addPanel() for each panel found in the config file. If there is
     * none, adds a new panel.
     * \param argc
     * \param argv
     */
    explicit NdePanelApplication(int& argc, char** argv);
    ~NdePanelApplication();

    void setIconTheme(const QString &iconTheme);

    /*!
     * \brief Determines the number of NdePanel objects
     * \return the current number of NdePanel objects
     */
    int count() const { return mPanels.count(); }

    /*!
     * \brief Checks if a given Plugin is running and has the
     * INdePanelPlugin::SingleInstance flag set. As Plugins are added to
     * NdePanel instances, this method only iterates over these NdePanel
     * instances and lets them check the conditions.
     * \param pluginId Plugin Identifier which is the basename of the .desktop
     * file that specifies the plugin.
     * \return true if the Plugin is running and has the
     * INdePanelPlugin::SingleInstance flag set, false otherwise.
     */
    bool isPluginSingletonAndRunnig(QString const & pluginId) const;

public Q_SLOTS:
    /*!
     * \brief Adds a new NdePanel which consists of the following steps:
     * 1. Create id/name.
     * 2. Create the NdePanel: call addPanel(name).
     * 3. Update the config file (add the new panel id to the list of panels).
     * 4. Show the panel configuration dialog so that the user can add plugins.
     *
     * This method will create a new NdePanel with a new name and add this
     * to the config file. So this should only be used while the application
     * is running and the user decides to add a new panel. At application
     * startup, addPanel() should be used instead.
     *
     * \note This slot will be used from the NdePanel right-click menu. As we
     * can only add new panels from a visible panel, we should never run
     * nde-panel without an NdePanel. Without a panel, we have just an
     * invisible application.
     */
    void addNewPanel();

    void addPanelFromSystemsettings(QString name);
    void removePanelFromSystemsettings(QString name);

Q_SIGNALS:
    /*!
     * \brief Signal that re-emits the signal pluginAdded() from NdePanel.
     */
    void pluginAdded();
    /*!
     * \brief Signal that re-emits the signal pluginRemoved() from NdePanel.
     */
    void pluginRemoved();

private:
    /*!
     * \brief Holds all the instances of NdePanel.
     */
    QList<NdePanel*> mPanels;
    /*!
     * \brief The global icon theme used by all apps (except for panels perhaps).
     */
    QString mGlobalIconTheme;
    /*!
     * \brief Creates a new NdePanel with the given name and connects the
     * appropriate signals and slots.
     * This method can be used at application startup.
     * \param name Name of the NdePanel as it is used in the config file.
     * \return The newly created NdePanel.
     */
    NdePanel* addPanel(const QString &name);

private Q_SLOTS:
    /*!
     * \brief Removes the given NdePanel which consists of the following
     * steps:
     * 1. Remove the panel from mPanels.
     * 2. Remove the panel from the config file.
     * 3. Schedule the QObject for deletion: QObject::deleteLater().
     * \param panel NdePanel instance that should be removed.
     */
    void removePanel(NdePanel* panel);

    /*!
     * \brief Connects the QScreen::destroyed signal of a new screen to
     * the screenDestroyed() slot so that we can handle this screens'
     * destruction as soon as it happens.
     * \param newScreen The QScreen that was created and added.
     */
    void handleScreenAdded(QScreen* newScreen);
    /*!
     * \brief Handles screen destruction. This is a workaround for a Qt bug.
     * For further information, see the implementation notes.
     * \param screenObj The QScreen that was destroyed.
     */
    void screenDestroyed(QObject* screenObj);
    /*!
     * \brief Reloads the panels. This is the second part of the workaround
     * mentioned above.
     */
    void reloadPanelsAsNeeded();
    /*!
     * \brief Deletes all NdePanel instances that are stored in mPanels.
     */
    void cleanup();
    /*!
     * \brief 双屏显示任务栏，当关闭其中一个屏幕时对应的任务栏也需要关闭
     */
    void onConfigurationChanged();

private:
    /*!
     * \brief mSettings is the Nde::Settings object that is used for the
     * current instance of nde-panel. Normally, this refers to the config file
     * $HOME/.config/nde/panel.conf (on Unix systems). This behaviour can be
     * changed with the -c command line option.
     */

    NdePanelApplicationPrivate *const d_ptr;

    Q_DECLARE_PRIVATE(NdePanelApplication)
    Q_DISABLE_COPY(NdePanelApplication)
};


#endif // NDEPANELAPPLICATION_H
