#include "desktopitem.h"
#include <QDBusReply>
#include <QDBusInterface>
#include <KWindowSystem/NETWM>
#include <QX11Info>

namespace NewStart {

QString argumentToString(const QVariant &arg)
{
    QString out;
    variantToString(arg, out);
    return out;
}

bool argToString(const QDBusArgument &busArg, QString &out)
{
    QString busSig = busArg.currentSignature();
    bool doIterate = false;
    QDBusArgument::ElementType elementType = busArg.currentType();

    if (elementType != QDBusArgument::BasicType && elementType != QDBusArgument::VariantType
            && elementType != QDBusArgument::MapEntryType)
        out += QLatin1String("[Argument: ") + busSig + QLatin1Char(' ');

    switch (elementType) {
        case QDBusArgument::BasicType:
        case QDBusArgument::VariantType:
            if (!variantToString(busArg.asVariant(), out))
                return false;
            break;
        case QDBusArgument::StructureType:
            busArg.beginStructure();
            doIterate = true;
            break;
        case QDBusArgument::ArrayType:
            busArg.beginArray();
            out += QLatin1Char('{');
            doIterate = true;
            break;
        case QDBusArgument::MapType:
            busArg.beginMap();
            out += QLatin1Char('{');
            doIterate = true;
            break;
        case QDBusArgument::MapEntryType:
            busArg.beginMapEntry();
            if (!variantToString(busArg.asVariant(), out))
                return false;
            out += QLatin1String(" = ");
            if (!argToString(busArg, out))
                return false;
            busArg.endMapEntry();
            break;
        case QDBusArgument::UnknownType:
        default:
            out += QLatin1String("<ERROR - Unknown Type>");
            return false;
    }
    if (doIterate && !busArg.atEnd()) {
        while (!busArg.atEnd()) {
            if (!argToString(busArg, out))
                return false;
            out += QLatin1String(", ");
        }
        out.chop(2);
    }
    switch (elementType) {
        case QDBusArgument::BasicType:
        case QDBusArgument::VariantType:
        case QDBusArgument::UnknownType:
        case QDBusArgument::MapEntryType:
            // nothing to do
            break;
        case QDBusArgument::StructureType:
            busArg.endStructure();
            break;
        case QDBusArgument::ArrayType:
            out += QLatin1Char('}');
            busArg.endArray();
            break;
        case QDBusArgument::MapType:
            out += QLatin1Char('}');
            busArg.endMap();
            break;
    }

    if (elementType != QDBusArgument::BasicType && elementType != QDBusArgument::VariantType
            && elementType != QDBusArgument::MapEntryType)
        out += QLatin1Char(']');

    return true;
}

bool variantToString(const QVariant &arg, QString &out)
{
    int argType = arg.userType();

    if (argType == QVariant::StringList) {
        out += QLatin1Char('{');
        const QStringList list = arg.toStringList();
        for (const QString &item : list)
            out += QLatin1Char('\"') + item + QLatin1String("\", ");
        if (!list.isEmpty())
            out.chop(2);
        out += QLatin1Char('}');
    } else if (argType == QVariant::ByteArray) {
        out += QLatin1Char('{');
        QByteArray list = arg.toByteArray();
        for (int i = 0; i < list.count(); ++i) {
            out += QString::number(list.at(i));
            out += QLatin1String(", ");
        }
        if (!list.isEmpty())
            out.chop(2);
        out += QLatin1Char('}');
    } else if (argType == QVariant::List) {
        out += QLatin1Char('{');
        const QList<QVariant> list = arg.toList();
        for (const QVariant &item : list) {
            if (!variantToString(item, out))
                return false;
            out += QLatin1String(", ");
        }
        if (!list.isEmpty())
            out.chop(2);
        out += QLatin1Char('}');
    } else if (argType == QMetaType::Char || argType == QMetaType::Short || argType == QMetaType::Int
               || argType == QMetaType::Long || argType == QMetaType::LongLong) {
        out += QString::number(arg.toLongLong());
    } else if (argType == QMetaType::UChar || argType == QMetaType::UShort || argType == QMetaType::UInt
               || argType == QMetaType::ULong || argType == QMetaType::ULongLong) {
        out += QString::number(arg.toULongLong());
    } else if (argType == QMetaType::Double) {
        out += QString::number(arg.toDouble());
    } else if (argType == QMetaType::Bool) {
        out += QLatin1String(arg.toBool() ? "true" : "false");
    } else if (argType == qMetaTypeId<QDBusArgument>()) {
        argToString(qvariant_cast<QDBusArgument>(arg), out);
    } else if (argType == qMetaTypeId<QDBusObjectPath>()) {
        const QString path = qvariant_cast<QDBusObjectPath>(arg).path();
        out += QLatin1String("[ObjectPath: ");
        out += path;
        out += QLatin1Char(']');
    } else if (argType == qMetaTypeId<QDBusSignature>()) {
        out += QLatin1String("[Signature: ") + qvariant_cast<QDBusSignature>(arg).signature();
        out += QLatin1Char(']');
    } else if (argType == qMetaTypeId<QDBusUnixFileDescriptor>()) {
        out += QLatin1String("[Unix FD: ");
        out += QLatin1String(qvariant_cast<QDBusUnixFileDescriptor>(arg).isValid() ? "valid" : "not valid");
        out += QLatin1Char(']');
    } else if (argType == qMetaTypeId<QDBusVariant>()) {
        const QVariant v = qvariant_cast<QDBusVariant>(arg).variant();
        out += QLatin1String("[Variant");
        int vUserType = v.userType();
        if (vUserType != qMetaTypeId<QDBusVariant>()
                && vUserType != qMetaTypeId<QDBusSignature>()
                && vUserType != qMetaTypeId<QDBusObjectPath>()
                && vUserType != qMetaTypeId<QDBusArgument>())
            out += QLatin1Char('(') + QLatin1String(v.typeName()) + QLatin1Char(')');
        out += QLatin1String(": ");
        if (!variantToString(v, out))
            return false;
        out += QLatin1Char(']');
    } else if (arg.canConvert(QVariant::String)) {
        out += QLatin1Char('\"') + arg.toString() + QLatin1Char('\"');
    } else {
        out += QLatin1Char('[');
        out += QLatin1String(arg.typeName());
        out += QLatin1Char(']');
    }

    return true;
}


windowitem::windowitem()
{

}

void windowitem::setwindow(WId id)
{
    window_id = id;
    //QImage imge = BackgroundManager::instance().getwindowimage(id);
    QImage imge  = ThumbnailManager::getIns()->getThumbnail(id).toImage();
    m_workplaceimg = imge;
    m_desktopimg = imge;
//    if(ThumbnailManager::getIns()->getThumbnail(id, pximap))
//    {
//        QImage imge = pximap.toImage();
//        m_workplaceimg = imge;
//        m_desktopimg = imge;
//    }
//    QImage imge;
//    if (BackgroundManager::instance().getwindowimage(id,imge)){
//        ImageManager::instance().addImage(id,imge);
//    } else if (ImageManager::instance().getImage(id,imge)){
//    } else {
//        imge = BackgroundManager::instance().getwindowimage(id);
//        ImageManager::instance().addImage(id,imge);
//    }
//    m_workplaceimg = imge;
//    m_desktopimg = imge;
}

void desktopitem::setDesktop(int d)
{
    if (cur_desktop != d)
    {
        cur_desktop = d;
        Q_EMIT desktopChanged();
    }
}

void desktopitem::deletewindow(WId id,bool close,bool refresh)
{
    if (close){
        KWinOpt::instance().removewindow(id);
    }

    auto iter = m_windowlist.find(id);
    if(iter != m_windowlist.end()){
        m_windowlist.remove(id);
        Q_EMIT backgrounddel(id);
        if (refresh) {
            Q_EMIT windowsChanged(cur_desktop);
        }

    }
}

void desktopitem::addwindow(QSharedPointer<windowitem> item)
{
    auto iter = m_windowlist.find(item->window());
    if(iter == m_windowlist.end()) {
        m_windowlist.insert(item->window(),item);
        Q_EMIT backgroundadd(item->window());
    }
}

void desktopitem::settitle(QString name)
{
    m_desktopname = name;
    Q_EMIT titleChanged();
}

QSharedPointer<windowitem> desktopitem::findwindowbywid(WId id)
{
    if(m_windowlist.find(id) != m_windowlist.end())
    {
        return m_windowlist[id];
    }

}

desktopitem::desktopitem()
{
    m_bgimg = ThumbnailManager::getIns()->getWallPaper();
    m_bgimg.scaled(m_rect.size(),Qt::KeepAspectRatioByExpanding,Qt::SmoothTransformation);
}

void desktopitem::initdesktop()
{
    for (auto wid: KWindowSystem::self()->windows())
    {
        KWindowInfo info(wid, NET::WMDesktop |NET::WMName);
        if (info.valid() && info.desktop() == cur_desktop)
        {
            if ((info.name().compare("desktopWidget") == 0)
               || (info.name().compare("") == 0)) {
                continue;
            }
            QSharedPointer<windowitem> iter(new windowitem());
            iter->setwindow(wid);
            iter->settitle(info.name());
            m_windowlist.insert(wid,iter);
        }
    }
}

void desktopmanager::Init()
{
    int number = KWinOpt::instance().getdesktopnum();
    for(int i = 1; i <= number; i++ )
    {
        QSharedPointer<desktopitem> item(new desktopitem());
        item->setDesktop(i);
        item->settitle(KWindowSystem::desktopName(i));
        item->initdesktop();
        m_managerlist.insert(i,item);
    }
}

void desktopmanager::deletedesktop(int desktop)
{
    ThumbnailManager::getIns()->onDesktopDelete(desktop);
    ThumbnailManager::getIns()->refreshScreenShot(KWinOpt::instance().getdesktopnum()-1);

    if (selectdesktopnum > desktop){
        //if (desktop != 1) {

            selectdesktopnum = selectdesktopnum -1;
        //}
    } else if (selectdesktopnum == desktop) {
        if(selectdesktopnum == KWinOpt::instance().getdesktopnum()){
            selectdesktopnum = selectdesktopnum -1;
        }
    }

    int _desktop = 0;
    for( auto it=m_managerlist.constBegin(); it!=m_managerlist.constEnd(); ++it)
    {
        if (it.key() != desktop){
            _desktop = it.key();
            break;
        }
    }
    auto windowlist = getdesktopbyid(desktop)->getwindowslist();

    for( auto it=windowlist.constBegin(); it!=windowlist.constEnd(); ++it)
    {
        KWinOpt::instance().removewindow(it.key());
    }

    //remove desktop
    if(KWinOpt::instance().removedesktop(desktop)){

    }


}

void desktopmanager::movewindowfromto(WId id,int from ,int to )
{
    KWindowSystem::setOnDesktop(id,to);
    QSharedPointer<windowitem> iter ;
    iter = m_managerlist[from]->findwindowbywid(id);
    m_managerlist[from]->deletewindow(id,false,false);
    m_managerlist[to]->addwindow(iter);
}

QSharedPointer<desktopitem> desktopmanager::getdesktopbyid(int id)
{
    if(m_managerlist.find(id) != m_managerlist.end()){
        return m_managerlist[id];
    }
    return nullptr;
}

void desktopmanager::removeitem(WId id,bool desktop)
{

}

void desktopmanager::adddesktop()
{
    KWinOpt::instance().createdesktop();
}

void KWinOpt::dumpMessage(const QDBusMessage &message)
{
    QList<QVariant> args = message.arguments();
    for (const QVariant &arg : qAsConst(args)) {
        QString str = argumentToString(arg).toHtmlEscaped();
        // turn object paths into clickable links
        str.replace(objectPathRegExp, QLatin1String("[ObjectPath: <a href=\"qdbus://bus\\1\">\\1</a>]"));
        // convert new lines from command to proper HTML line breaks
        str.replace(QStringLiteral("\n"), QStringLiteral("<br/>"));
        str.replace(QStringLiteral("&quot;"), QStringLiteral(""));
        str.replace(QStringLiteral("Argument: (uss) "), QStringLiteral(""));
        str.replace(QStringLiteral("[Variant: [Argument: a(uss) {"), QStringLiteral(""));
        str.replace(QStringLiteral("}]]"), QStringLiteral(""));

        QStringList lstr = str.split("], ");
        for(int i = 0; i< lstr.size();++i)
        {
            QString tmp = lstr.at(i);
            tmp.replace(QStringLiteral("["), QStringLiteral(""));
            tmp.replace(QStringLiteral("]"), QStringLiteral(""));
            QStringList item = tmp.split(",");
            if (item.size() == 3) {
            }
        }
    }
}

int KWinOpt::getdesktopnum()
{
    return  NETRootInfo(QX11Info::connection(), NET::NumberOfDesktops).numberOfDesktops();
//    return KWindowSystem::numberOfDesktops();
//    QDBusInterface intface("com.kde.kwin","/nde","com.kde.KWin",QDBusConnection::sessionBus());
//    QDBusReply<int> reply = intface.call("GetDesktopNumber");
//    int value = reply.value();
//    return value;
}

int KWinOpt::createdesktop() {

    NETRootInfo(QX11Info::connection(), NET::NumberOfDesktops).setNumberOfDesktops(getdesktopnum()+1);
    return 0;
//    KWindowSystem::();
//    QDBusMessage message = QDBusMessage::createMethodCall("com.kde.kwin",
//                           "/nde",
//                           "com.kde.KWin",
//                           "CreateDesktop");
//    QDBusMessage reply = QDBusConnection::sessionBus().call(message);
//    if (reply.type() == QDBusMessage::ReplyMessage){
//        return 0;
//    }
//    else{
//        return 1;
//    }
}

bool KWinOpt::removedesktop(int num){

    int count = this->getdesktopnum();

    if(num <= 0 || num > count || count == 1)
        return false;

    for(WId wid : KWindowSystem::stackingOrder())
    {
         if(KWindowInfo(wid, NET::WMDesktop).isOnDesktop(num))
             NETRootInfo(QX11Info::connection(), NET::CloseWindow).closeWindowRequest(wid);
    }
    NETRootInfo(QX11Info::connection(), NET::NumberOfDesktops).setNumberOfDesktops(getdesktopnum()-1);

    return true;

//    QDBusMessage message = QDBusMessage::createMethodCall("com.kde.kwin",
//                           "/nde",
//                           "com.kde.KWin",
//                           "removeDesktop");
//    QList<QVariant> args;
//    args.insert(0,num);
//    message.setArguments(args);
//    //发送消息
//    QDBusMessage response = QDBusConnection::sessionBus().call(message);

//    if (response.type() == QDBusMessage::ReplyMessage){
//        return true;
//    }
//    else{
//        return false;
//    }

}

bool KWinOpt::removewindow(WId id) {

    NETRootInfo(QX11Info::connection(), NET::CloseWindow).closeWindowRequest(id);
    return true;
//    QDBusMessage message = QDBusMessage::createMethodCall("com.kde.kwin",
//                           "/nde",
//                           "com.kde.KWin",
//                           "removeWindow");
//    QList<QVariant> args;
//    args.insert(0,QString::number(id));
//    message.setArguments(args);
//    //发送消息
//    QDBusMessage response = QDBusConnection::sessionBus().call(message);

//    if (response.type() == QDBusMessage::ReplyMessage){
//    }
//    else{
//    }
//    return true;
}

}
