#include "displaywidget.h"
#include <KWindowEffects>
#include <QBitmap>
#include <QPainter>
#include <QEvent>
#include <QKeyEvent>
#include <QApplication>
#include <QScreen>
#include <QBitmap>
#include <QDesktopServices>
#include <QDesktopWidget>
#include <QStyle>
#include "appmanager.h"
#include "../panel/thumbnailmanager.h"
#include <QPainterPath>

#include <QDebug>

DisplayWidget::DisplayWidget(INdePanelPlugin *plugin, QWidget *parent)
    : QWidget(parent),
      mPlugin(plugin),
      mCompositorOpen(false),
      mMenuDisplayMode(DisplayNormal)
{
    this->setObjectName("DisplayWidget");
    this->setWindowFlags(Qt::FramelessWindowHint |
                         Qt::WindowStaysOnTopHint |
                         Qt::Tool |
                         Qt::X11BypassWindowManagerHint);
    this->setAttribute(Qt::WA_TranslucentBackground);
    this->setAttribute(Qt::WA_NoSystemBackground, false);
    KWindowEffects::enableBlurBehind(this->winId(), true);

    mBackgroundWidget = new QWidget(this);
    mBackgroundWidget->setObjectName("DisplayBackgroundWidget");

    mDisplayMainMenu = new DisplayMainMenu(plugin, this);
    connect(mDisplayMainMenu, &DisplayMainMenu::signalAddToDesktop,this,&DisplayWidget::slotAddToDesktop);
    connect(mDisplayMainMenu, &DisplayMainMenu::signalAddToQuickLaunch,this,&DisplayWidget::slotAddToQuickLaunch);
    mDisplayMainMenu->hide();

    mDisplayMainMenuFull = new DisplayMainMenuFull(plugin, this);
    connect(mDisplayMainMenuFull, &DisplayMainMenuFull::signalAddToDesktop,this,&DisplayWidget::slotAddToDesktop);
    connect(mDisplayMainMenuFull, &DisplayMainMenuFull::signalAddToQuickLaunch,this,&DisplayWidget::slotAddToQuickLaunch);
    connect(AppManager::getIns(), &AppManager::menuDisplayModeChanged,this,&DisplayWidget::slotMainMenuModeSwitch);
    mDisplayMainMenuFull->hide();

    QHBoxLayout *bLayout = new QHBoxLayout;
    bLayout->setMargin(0);
    bLayout->setSpacing(0);
    bLayout->setContentsMargins(0,0,0,0);
    bLayout->addWidget(mDisplayMainMenu);
    bLayout->addWidget(mDisplayMainMenuFull);
    mBackgroundWidget->setLayout(bLayout);

    QHBoxLayout *layout = new QHBoxLayout;
    layout->setMargin(0);
    layout->setSpacing(0);
    layout->setContentsMargins(0,0,0,0);
    layout->addWidget(mBackgroundWidget);
    this->setLayout(layout);

    mAnimation = new QPropertyAnimation(this, "geometry");
    mAnimation->setEasingCurve(QEasingCurve::Linear);
    mAnimation->setDuration(100);
    connect(mAnimation, &QPropertyAnimation::finished, this, &DisplayWidget::slotModeSwitchAnimFinished);

    mCompositorOpen = mPlugin->panel()->compositorOpen();
    mBackgroundWidget->setProperty("compositor", mCompositorOpen);
    mBackgroundWidget->setProperty("full", (mMenuDisplayMode == DisplayFull));

    updateBackground();
}

void DisplayWidget::showMainMenu()
{
    AppManager::getIns()->restoreMenuDisplayMode();
    mMenuDisplayMode = AppManager::getIns()->menuDisplayMode();

    showMainMenuByMode();
}

void DisplayWidget::realign()
{
    mBackgroundWidget->setProperty("full", (mMenuDisplayMode == DisplayFull));

    mCompositorOpen = mPlugin->panel()->compositorOpen();
    mCurPanelPos = mPlugin->panel()->position();

    mBackgroundWidget->setProperty("compositor", mCompositorOpen);
    mBackgroundWidget->setProperty("direction", "");
    if(mMenuDisplayMode == DisplayNormal)
    {
        switch(mPlugin->panel()->position())
        {
        case NdePanel::PositionBottom:
            mBackgroundWidget->setProperty("direction", "top-right");
            break;
        case NdePanel::PositionTop:
        case NdePanel::PositionLeft:
            mBackgroundWidget->setProperty("direction", "bottom-right");
            break;
        case NdePanel::PositionRight:
            mBackgroundWidget->setProperty("direction", "bottom-left");
            break;
        }
    }
    mBackgroundWidget->style()->unpolish(mBackgroundWidget);
    mBackgroundWidget->style()->polish(mBackgroundWidget);
}

void DisplayWidget::slotMainMenuModeSwitch(MenuDisplayMode mode)
{
    if(mMenuDisplayMode == mode) return;

    if(mAnimation->state() == QPropertyAnimation::Running)
    {
        mAnimation->pause();
    }

    QRect oldRect = calGeometry();
    mMenuDisplayMode = mode;
    QRect newRect = calGeometry();
    mAnimation->setStartValue(oldRect);
    mAnimation->setEndValue(newRect);

    //若切换之后为小窗口模式，由于全屏界面为透明背景，导致动画看不见，需要先设置界面属性。
    if(mMenuDisplayMode == DisplayNormal)
    {
        realign();
    }

    //animation start
    mDisplayMainMenu->hide();
    mDisplayMainMenuFull->hide();
    mAnimation->start();
}

void DisplayWidget::slotModeSwitchAnimFinished()
{
    showMainMenuByMode();
}

void DisplayWidget::slotAddToDesktop(XdgDesktopFile desktopFile)
{
    QString desktopPath = QStandardPaths::writableLocation(QStandardPaths::DesktopLocation);
    //直接保存destop文件到桌面可能会遇到无法打开的问题。
    //desktopFile.save(desktopPath +"/"+ XdgDesktopFile::id( desktopFile.fileName(),false));
    QString sourceFile = desktopFile.fileName();
    QString destFile = desktopPath +"/";
    if(sourceFile.contains(" ")) // 解决文件名有空格时无法复制问题
    {
        QString path = sourceFile.left(sourceFile.lastIndexOf("/"));
        sourceFile = path + "/\'" + XdgDesktopFile::id( desktopFile.fileName()) + "\'";
    }
    
    QString tmp = QString("cp %1 %2").arg(sourceFile).arg(destFile);
    QStringList args;
    args << "-c" << tmp.toLocal8Bit().data();
    QProcess::startDetached("/bin/bash", args);
}

void DisplayWidget::slotAddToQuickLaunch(XdgDesktopFile desktopFile)
{
    NdePanel *panel = static_cast<NdePanel *>( mPlugin->panel() );
    if(panel)
    {
        const INdePanelPlugin * c_i_plugin =  panel->getPlugin("QuickLaunch");
        INdePanelPlugin * i_plugin = const_cast<INdePanelPlugin *>(c_i_plugin);
        NdeQuickLaunchPlugin * plugin = dynamic_cast<NdeQuickLaunchPlugin *>(i_plugin);
        if(plugin)
        {
            NdeQuickLaunch * widget = qobject_cast<NdeQuickLaunch*>( plugin->widget());
            if(widget)
            {
                widget->addButtonQuick(desktopFile);
            }
        }
    }
}

bool DisplayWidget::event(QEvent *event)
{
    if(event->type() == QEvent::KeyPress ||
            event->type() == QEvent::KeyRelease)
    {
        QKeyEvent* ke = static_cast<QKeyEvent*>(event);
        if(ke && ke->key() == Qt::Key_Escape)
        {
            this->hide();
        }
    }
    if(event->type() == QEvent::ActivationChange)
    {
        if(!this->isActiveWindow())
        {
            //这里需要延时隐藏界面，由于快捷键一按下就会触发此事件，界面就会出现闪烁。
            QTimer::singleShot(200, this, [this](){this->hide();});
        }
    }
    return QWidget::event(event);
}

void DisplayWidget::resizeEvent(QResizeEvent *event)
{
    QWidget::resizeEvent(event);
}

void DisplayWidget::paintEvent(QPaintEvent *event)
{
    QWidget::paintEvent(event);

    this->clearMask();
    if(mMenuDisplayMode == DisplayFull)
    {
        QPainter p(this);
        p.drawPixmap(this->rect(), mBlurBackground.scaled(this->size()));
    }
    else
    {
        QRect rect = this->rect().adjusted(0,0,0,0);

        QBitmap bmp(rect.size());

        bmp.fill();

        QPainter p(&bmp);

        //这里随便设置颜色，不会影响到qss的背景色效果
        p.setBrush(Qt::black);
        p.setRenderHint(QPainter::SmoothPixmapTransform);

        int arcR = 11; //圆弧大小
        QPainterPath path;
        switch(mPlugin->panel()->position())
        {
        case NdePanel::PositionBottom:
            path.moveTo(0, 0);
            path.lineTo(0, rect.height());
            path.lineTo(rect.width(), rect.height());
            path.lineTo(rect.width(), arcR);
            path.arcTo(rect.width() - arcR*2, 0, arcR * 2, arcR * 2, 0, 90);
            path.lineTo(0,0);

            break;
        case NdePanel::PositionTop:
        case NdePanel::PositionLeft:
            path.moveTo(0, 0);
            path.lineTo(0, rect.height());
            path.lineTo(rect.width() - arcR, rect.height());
            path.arcTo(rect.width() - arcR*2, rect.height() - arcR*2, arcR * 2, arcR * 2, 270, 90);
            path.lineTo(rect.width(), 0);
            path.lineTo(0,0);

            break;
        case NdePanel::PositionRight:
            path.moveTo(0, 0);
            path.lineTo(0, rect.height() - arcR);
            path.arcTo(0, rect.height() - arcR*2, arcR * 2, arcR * 2, 180, 90);
            path.lineTo(rect.width(), rect.height());
            path.lineTo(rect.width(), 0);
            path.lineTo(0,0);
        }

        p.drawPath(path);
        this->setMask(bmp);
    }
}

void DisplayWidget::showMainMenuByMode()
{
    updateBackground();
    realign();

    this->setGeometry(calGeometry());
    this->show();

    if(mMenuDisplayMode == DisplayNormal)
    {
        mDisplayMainMenuFull->hide();
        mDisplayMainMenu->show();
        mDisplayMainMenu->restoreUI();
    }
    else if(mMenuDisplayMode == DisplayFull)
    {
        mDisplayMainMenu->hide();
        mDisplayMainMenuFull->show();
        mDisplayMainMenuFull->restoreUI();
    }
}

QRect DisplayWidget::calGeometry()
{
    QRect rect;
    QRect availableRect = qApp->primaryScreen()->availableGeometry();

    if(mMenuDisplayMode == DisplayFull)
    {
        rect = availableRect;
    }
    else
    {
        QSize fixedSize = QSize(qMin(availableRect.width(), 374)  ,qMin(availableRect.height(), 458));
        rect = mPlugin->calculatePopupWindowPos(fixedSize);
    }
    return rect;
}

void DisplayWidget::updateBackground()
{
    mBlurBackground = ThumbnailManager::getIns()->getAdjustWallPaper();

    QSize desktopsize(QApplication::desktop()->screenGeometry().width(), QApplication::desktop()->screenGeometry().height());
    mBlurBackground.scaled(desktopsize, Qt::KeepAspectRatioByExpanding,Qt::SmoothTransformation);
}
