/* BEGIN_COMMON_COPYRIGHT_HEADER
 * (c)LGPL2+
 *
 * Nde - a lightweight, Qt based, desktop toolset
 * https://lxqt.org
 *
 * Copyright: 2010-2011 Razor team
 * Authors:
 *   Alexander Sokoloff <sokoloff.a@gmail.com>
 *
 * This program or library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301 USA
 *
 * END_COMMON_COPYRIGHT_HEADER */


#include "ndemainmenu.h"
#include "ndemainmenuconfiguration.h"
#include "../panel/ndepanel.h"
#include <QTimer>
#include <QMessageBox>
#include <QEvent>
#include <QKeyEvent>
#include <QResizeEvent>
#include <nde-globalkeys.h>
#include <QApplication>
#include <QDebug>
#include <XdgIcon>
#include <QScreen>
#include "appmanager.h"

#define DEFAULT_SHORTCUT1 "Super_L"
#define DEFAULT_SHORTCUT2 "Control+Escape"

NdeMainMenu::NdeMainMenu(const INdePanelPluginStartupInfo &startupInfo):
    QObject(),
    INdePanelPlugin(startupInfo),
    mShortcut1(nullptr),
    mShortcut2(nullptr)
{
    //主界面
    mDisplay = new DisplayWidget(this);
    //panel位置
    mCurPanelPos = this->panel()->position();

    mButton.setAutoRaise(true);
    //Notes:
    //1. installing event filter to parent widget to avoid infinite loop
    //   (while setting icon we also need to set the style)
    //2. delaying of installEventFilter because in c-tor mButton has no parent widget
    //   (parent is assigned in panel's logic after widget() call)

    connect(&mButton, &QToolButton::clicked, this, &NdeMainMenu::showHideMenu);

    mShortcut1 = GlobalKeyShortcut::Client::instance()->addAction(
                QString{}, QString("/panel/%1/show_hide1").arg(settings()->group()),
                NdeMainMenu::tr("Show/hide main menu") + "1", this);
    if (mShortcut1)
    {
        connect(mShortcut1, &GlobalKeyShortcut::Action::registrationFinished, [this] {
            if (mShortcut1->shortcut().isEmpty())
                mShortcut1->changeShortcut(DEFAULT_SHORTCUT1);
        });
        connect(mShortcut1, &GlobalKeyShortcut::Action::activated, [this] {
            showHideMenu();
        });
    }
    mShortcut2 = GlobalKeyShortcut::Client::instance()->addAction(
                QString{}, QString("/panel/%1/show_hide2").arg(settings()->group()),
                NdeMainMenu::tr("Show/hide main menu") + "2", this);
    if (mShortcut2)
    {
        connect(mShortcut2, &GlobalKeyShortcut::Action::registrationFinished, [this] {
            if (mShortcut2->shortcut().isEmpty())
                mShortcut2->changeShortcut(DEFAULT_SHORTCUT2);
        });
        connect(mShortcut2, &GlobalKeyShortcut::Action::activated, [this] {
            showHideMenu();
        });
    }

    QDBusConnection connect = QDBusConnection::sessionBus();
    if(!connect.registerService("com.nde.panel.mainmenu"))
    {
        //
        qWarning() << "D-BUS register service failed:com.nde.panel.mainmenu";
    }
    if(!connect.registerObject("/com/nde/panel/mainmenu", AppManager::getIns(),QDBusConnection::ExportAllSlots))
    {
        //
        qWarning() << "D-BUS register object failed:/com/nde/panel/mainmenu";
    }

    QTimer::singleShot(0, [this] { Q_ASSERT(mButton.parentWidget()); mButton.parentWidget()->installEventFilter(this); });
    QTimer::singleShot(0, [this] { readXDGMenuFile(); });
}

NdeMainMenu::~NdeMainMenu()
{
    mButton.parentWidget()->removeEventFilter(this);
}

void NdeMainMenu::showHideMenu()
{
    if(!mDisplay->isActiveWindow()) {
        panel()->willShowWindow(mDisplay);
        mDisplay->showMainMenu();
        mDisplay->activateWindow();
    }
    else
    {
        mDisplay->hide();
    }
}

void NdeMainMenu::readXDGMenuFile()
{
    setButtonIcon();
    if (settings()->value("showText", false).toBool())
    {
        mButton.setText(settings()->value("text", "Start").toString());
        mButton.setToolButtonStyle(Qt::ToolButtonTextBesideIcon);
    }
    else
    {
        mButton.setText("");
        mButton.setToolButtonStyle(Qt::ToolButtonIconOnly);
    }

    mLogDir = settings()->value("log_dir", "").toString();

    QString menu_file = settings()->value("menu_file", "").toString();
    if (menu_file.isEmpty())
        menu_file = XdgMenu::getMenuFileName();

    if (mMenuFile != menu_file)
    {
        mMenuFile = menu_file;
        mXdgMenu.setEnvironments(QStringList() << "X-NDE" << "Nde");
        mXdgMenu.setLogDir(mLogDir);

        bool res = mXdgMenu.read(mMenuFile);
        qWarning() << "mMenuFile::" << mMenuFile;
        connect(&mXdgMenu, &XdgMenu::changed, [this](){
            QTimer::singleShot(1000, this, &NdeMainMenu::buildMenu);
        });
        if (res)
        {
            QTimer::singleShot(1000, this, &NdeMainMenu::buildMenu);
        }
        else
        {
            QMessageBox::warning(0, "Parse error", mXdgMenu.errorString());
            return;
        }
    }
    realign();
}

/************************************************

 ************************************************/
void NdeMainMenu::buildMenu()
{
    AppManager::getIns()->rebuildMenu(mXdgMenu.xml());
//    mDisplay->rebuildMenu(mXdgMenu.xml());

    //快速启动栏重新加载
    NdePanel *panel = static_cast<NdePanel *>( this->panel() );
    if(panel)
    {
        const INdePanelPlugin * c_i_plugin =  panel->getPlugin("QuickLaunch");
        INdePanelPlugin * i_plugin = const_cast<INdePanelPlugin *>(c_i_plugin);
        NdeQuickLaunchPlugin * plugin = dynamic_cast<NdeQuickLaunchPlugin *>(i_plugin);
        if(plugin)
        {
            NdeQuickLaunch * widget = qobject_cast<NdeQuickLaunch*>( plugin->widget());
            if(widget)
            {
                widget->rebuildButtonQuick();
            }
        }
    }
}

void NdeMainMenu::setButtonIcon()
{
    if (settings()->value("ownIcon", false).toBool())
    {
        mButton.setStyleSheet(QString("#MainMenu { qproperty-icon: url(%1); }")
                .arg(settings()->value(QLatin1String("icon"), QLatin1String(NDE_GRAPHICS_DIR"/helix.svg")).toString()));
    } else
    {
        mButton.setStyleSheet(QString());
    }
}

//QDialog *NdeMainMenu::configureDialog()
//{
//    return new NdeMainMenuConfiguration(settings(), mShortcut, DEFAULT_SHORTCUT);
//}

void NdeMainMenu::realign()
{
    mDisplay->realign();

    switch(this->panel()->position())
    {
    case NdePanel::PositionBottom:
    case NdePanel::PositionTop:
    {
        int iconSize = this->panel()->iconSize();
        mButton.setMinimumHeight(iconSize);
        mButton.setMaximumHeight(iconSize+100);
        mButton.setFixedWidth(58);
        mButton.setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Minimum);
        mButton.setIconSize(QSize(iconSize, iconSize));
        break;
    }
    case NdePanel::PositionLeft:
    case NdePanel::PositionRight:
    {
        int iconSize = this->panel()->iconSize();
        mButton.setMinimumWidth(iconSize);
        mButton.setMaximumWidth(iconSize+100);
        mButton.setFixedHeight(58);
        mButton.setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Expanding);
        mButton.setIconSize(QSize(iconSize, iconSize));
        break;
    }
    }
    //panel位置变化,或者屏幕尺寸变化，隐藏掉popup
    INdePanel::Position pos = this->panel()->position();
    QRect rc = qApp->primaryScreen()->geometry();

    if(mDisplay && (mCurPanelPos != pos || mScreenRect != rc))
    {
        mDisplay->hide();
        mCurPanelPos = pos;
        mScreenRect = rc;
    }
}

// functor used to match a QAction by prefix
struct MatchAction
{
    MatchAction(QString key):key_(key) {}
    bool operator()(QAction* action) { return action->text().startsWith(key_, Qt::CaseInsensitive); }
    QString key_;
};

bool NdeMainMenu::eventFilter(QObject *obj, QEvent *event)
{
    if(obj == mButton.parentWidget())
    {
        // the application is given a new QStyle
        if(event->type() == QEvent::StyleChange)
        {
            setButtonIcon();
        }
    }
    return false;
}

#undef DEFAULT_SHORTCUT
