#include "notificationbar.h"
#include "notificationsadaptor.h"
#include "notifyd.h"
#include <QStyle>
#include <QMouseEvent>
#include <QHBoxLayout>
#include <QScreen>
#include <QtDBus/QDBusConnection>

#include <QDebug>

NotificationBarWidget::NotificationBarWidget(INdePanelPlugin *plugin, QWidget *parent) :
    QFrame(parent),
    mPlugin(plugin)
{
    mDisplayWidth = NOTIFICATIONBAR_WIDTH;
    mIsDisplayShow = false;

    QHBoxLayout *layout = new QHBoxLayout(this);
    layout->setContentsMargins (0, 0, 0, 0);

    //btn
    mBtn = new QToolButton(this);
    mBtn->setObjectName("NotificationBarWidgetBtn");
    mBtn->setProperty("message", false);
    mBtn->setAutoRaise(true);
    mBtn->setSizePolicy(QSizePolicy::Expanding,QSizePolicy::Expanding);
    connect(mBtn, &QToolButton::clicked, this, &NotificationBarWidget::onBtnClicked);

    mPopup = new NotifyPopup(plugin);
    mPopup->hide();
    connect(mPopup, &NotifyPopup::triggered, this, &NotificationBarWidget::onPopupTriggered);

    //default
    QRect rc = qApp->primaryScreen()->availableGeometry();
    mDisplay = new NotifyDisplayWidget(plugin);
    mDisplay->setGeometry(rc.adjusted(rc.width(),0,mDisplayWidth,0));
    mDisplay->hide();

    connect(mDisplay, &NotifyDisplayWidget::lostFocus,
            this, &NotificationBarWidget::onDisplayLostFocus);
    connect(mDisplay, &NotifyDisplayWidget::actionNewNotify,
            this, &NotificationBarWidget::onNotifiesCountChanged);

    //animation
    mAnimation = new QPropertyAnimation(mDisplay, "geometry", mDisplay);
    mAnimation->setDuration(200);
    mAnimation->setEasingCurve(QEasingCurve::Linear);
    mAnimation->setStartValue(rc.adjusted(rc.width(),0,mDisplayWidth,0));
    mAnimation->setEndValue(rc.adjusted(rc.width()-mDisplayWidth,0,0,0));
    connect(mAnimation, &QPropertyAnimation::finished, this, &NotificationBarWidget::onAnimationFinished);

    layout->addWidget(mBtn);
    setLayout(layout);

    //add dbus service
    Notifyd* daemon = new Notifyd(mDisplay, mPopup, this);
    new NotificationsAdaptor(daemon);

    QDBusConnection connection = QDBusConnection::sessionBus();
    if (!connection.registerService(QSL("org.freedesktop.Notifications")))
        qDebug() << "registerService failed: another service with 'org.freedesktop.Notifications' runs already";
    if (!connection.registerObject(QSL("/org/freedesktop/Notifications"), daemon))
        qDebug() << "registerObject failed: another object with '/org/freedesktop/Notifications' runs already";
}

NotificationBarWidget::~NotificationBarWidget()
{
    QDBusConnection::sessionBus().unregisterService("org.freedesktop.Notifications");

    mAnimation->deleteLater();
    mDisplay->deleteLater();
    mBtn->deleteLater();
}

void NotificationBarWidget::realign()
{
    mPopup->realign();
    mDisplay->realign();

    NdePanel *panel = static_cast<NdePanel *>( mPlugin->panel() );
    if(panel)
    {
        INdePanel::Position pos = panel->position();
        QRect rc = qApp->primaryScreen()->geometry();
        //位置变化或者屏幕尺寸变化隐藏该控件
        if(pos != mPanelPos || mScreenRect != rc)
        {
            mPanelPos = pos;
            mScreenRect = rc;

            switch(pos)
            {
            case INdePanel::PositionTop:
                rc.adjust(0,panel->panelSize(),0,0);
                break;
            case INdePanel::PositionBottom:
                rc.adjust(0,0,0,-panel->panelSize());
                break;
            case INdePanel::PositionLeft:
                rc.adjust(0,0,0,0);
                break;
            case INdePanel::PositionRight:
                rc.adjust(0,0,-panel->panelSize(),0);
                break;
            }

            mAnimation->setStartValue(rc.adjusted(rc.width(),0,mDisplayWidth,0));
            mAnimation->setEndValue(rc.adjusted(rc.width()-mDisplayWidth,0,0,0));

            mIsDisplayShow = false;
            mDisplay->hide();

            mPopup->hide();
        }
    }
}

void NotificationBarWidget::onDisplayLostFocus()
{
    if(mIsDisplayShow)
    {
        mIsDisplayShow = !mIsDisplayShow;
        if(mAnimation->state() == QPropertyAnimation::Running)
            mAnimation->pause();
        QPropertyAnimation::Direction d = mIsDisplayShow ?
                    QPropertyAnimation::Forward:QPropertyAnimation::Backward;

        mAnimation->setDirection(d);
        mAnimation->start();
    }
}

void NotificationBarWidget::onNotifiesCountChanged()
{
    mBtn->setProperty("message", true);
    mBtn->style()->unpolish(mBtn);
    mBtn->style()->polish(mBtn);
}

void NotificationBarWidget::onBtnClicked()
{
    mIsDisplayShow = !mIsDisplayShow;

    if(mAnimation->state() == QPropertyAnimation::Running)
        mAnimation->pause();
    QPropertyAnimation::Direction d = mIsDisplayShow ?
                QPropertyAnimation::Forward:QPropertyAnimation::Backward;

    mAnimation->setDirection(d);

    //
    if(mIsDisplayShow)
    {
        mPlugin->panel()->willShowWindow(mDisplay);
        mDisplay->show();
        mDisplay->activateWindow();
    }

    mAnimation->start();
    mBtn->setProperty("message", false);
    mBtn->style()->unpolish(mBtn);
    mBtn->style()->polish(mBtn);
}

void NotificationBarWidget::onAnimationFinished()
{
    mDisplay->setVisible(mIsDisplayShow);
}

void NotificationBarWidget::onPopupTriggered()
{

}

//-------------------------------

NotificationBar::NotificationBar(
        const INdePanelPluginStartupInfo &startupInfo)
    : QObject(),
      INdePanelPlugin(startupInfo),
      mVisible(true)
{
    mWidget = new NotificationBarWidget(this);
    settingsChanged();
    realign();
}

NotificationBar::~NotificationBar()
{

}

void NotificationBar::realign()
{
    mWidget->realign();
}

void NotificationBar::settingsChanged()
{
    bool visible = this->settings()->value("visible","true").toBool();
    if(mVisible != visible)
    {
        mWidget->setVisible(visible);
        mVisible = visible;
    }
}
