#include "notifymessage.h"
#include <QDBusArgument>
#include <QUrl>
#include <QFile>
#include <XdgIcon>

void NotifyMessage::setMessage(
        const QString &app_name,
        uint replaces_id,
        const QString &app_icon,
        const QString &summary,
        const QString &body,
        const QStringList &actions,
        const QVariantMap &hints,
        int expire_timeout,
        bool noSave)
{
    m_id = replaces_id;
    m_name = app_name;
    m_summary = summary;

    //详细描述
    //https://developer.gnome.org/notification-spec
    //Body - This is a multi-line body of text. Each line is a paragraph, server implementations are free to word wrap them as they see fit.
    //XXX: remove all unsupported tags?!? (supported <b>, <i>, <u>, <a>, <img>)
    m_body = body;
    m_timeout = expire_timeout;
    m_noSave = noSave;

    m_origin_icon = app_icon;
    m_origin_actions = actions;
    m_origin_hints = hints;

    //m_icon
    //图标
    if (!app_icon.isEmpty())
    {
        m_icon = getPixmapFromString(app_icon);
    }
    else if (!hints[QLatin1String("icon_data")].isNull())
    {
        m_icon = getPixmapFromHint(hints[QLatin1String("icon_data")]);
    }

    //图标不存在，读取默认图标
    if (m_icon.isNull())
    {
        m_icon = XdgIcon::fromTheme(QLatin1String("application-x-executable")).pixmap(NOTIFY_ICON_SIZE);
    }
    else
    {
        if (m_icon.size().width() > NOTIFY_ICON_SIZE.width()
                || m_icon.size().height() > NOTIFY_ICON_SIZE.height())
        {
            m_icon = m_icon.scaled(NOTIFY_ICON_SIZE, Qt::KeepAspectRatio, Qt::SmoothTransformation);
        }
    }

    //m_pixmap
    //图片
    if (!hints[QLatin1String("image_data")].isNull())
    {
        m_pixmap = getPixmapFromHint(hints[QLatin1String("image_data")]);
    }
    else if (!hints[QLatin1String("image_path")].isNull())
    {
        m_pixmap = getPixmapFromString(hints[QLatin1String("image_path")].toString());
    }

    //m_actions
    for (int i = 0; i < actions.count(); i += 2)
    {
        QString key = actions[i];
        QString desc;
        if(i == actions.count() -1) {
            desc = key;
        }else{
            desc = actions[i+1];
        }
        m_actions.append(QPair<QString, QString>(key, desc));
    }

    // Categories *********************
//    if (!hints[QLatin1String("category")].isNull())
//    {
//        // TODO/FIXME: Categories - how to handle it?
//    }

    // Urgency Levels *********************
    // Type    Description
    // 0   Low
    // 1   Normal
    // 2   Critical
//    if (!hints[QLatin1String("urgency")].isNull())
//    {
//        // TODO/FIXME: Urgencies - how to handle it?
//    }
}

QPixmap NotifyMessage::getPixmapFromHint(const QVariant &argument) const
{
    int width, height, rowstride, bitsPerSample, channels;
    bool hasAlpha;
    QByteArray data;

    const QDBusArgument arg = argument.value<QDBusArgument>();
    arg.beginStructure();
    arg >> width;
    arg >> height;
    arg >> rowstride;
    arg >> hasAlpha;
    arg >> bitsPerSample;
    arg >> channels;
    arg >> data;
    arg.endStructure();

    bool rgb = !hasAlpha && channels == 3 && bitsPerSample == 8;
    QImage::Format imageFormat = rgb ? QImage::Format_RGB888 : QImage::Format_ARGB32;

    QImage img = QImage((uchar*)data.constData(), width, height, imageFormat);

    if (!rgb)
        img = img.rgbSwapped();

    return QPixmap::fromImage(img);
}

QPixmap NotifyMessage::getPixmapFromString(const QString &str) const
{
    QUrl url(str);
    if (url.isValid() && QFile::exists(url.toLocalFile()))
    {
        return QPixmap(url.toLocalFile());
    }
    else
    {
        return XdgIcon::fromTheme(str/*, XdgIcon::defaultApplicationIcon()*/).pixmap(NOTIFY_ICON_SIZE);
    }
}
