/* BEGIN_COMMON_COPYRIGHT_HEADER
 * (c)LGPL2+
 *
 * Nde - a lightweight, Qt based, desktop toolset
 * https://lxqt.org
 *
 * Copyright: 2010-2011 Razor team
 * Authors:
 *   Petr Vanek <petr@scribus.info>
 *
 * This program or library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301 USA
 *
 * END_COMMON_COPYRIGHT_HEADER */

#include "quicklaunchbutton.h"
#include "ndequicklaunch.h"
#include "../panel/indepanelplugin.h"
#include <QAction>
#include <QDrag>
#include <QMenu>
#include <QMouseEvent>
#include <QStylePainter>
#include <QStyleOptionToolButton>
#include <QApplication>
#include <XdgIcon>
#include <QDebug>
#include <QGraphicsDropShadowEffect>

#define MIMETYPE "x-nde/quicklaunch-button"


QuickLaunchButton::QuickLaunchButton(QuickLaunchAction * act, INdePanelPlugin * plugin, NdeQuickLaunch *launch, QWidget * parent)
    : QToolButton(parent),
      mAct(act),
      mPlugin(plugin),
      mLaunch(launch),
      mDragStart(false)
{
    setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);
    setAcceptDrops(true);
    setAutoRaise(true);

    setDefaultAction(mAct);
    mAct->setParent(this);

    mMoveLeftAct = new QAction(XdgIcon::fromTheme("go-previous"), tr("Move left"), this);
    connect(mMoveLeftAct, SIGNAL(triggered()), this, SIGNAL(movedLeft()));

    mMoveRightAct = new QAction(XdgIcon::fromTheme("go-next"), tr("Move right"), this);
    connect(mMoveRightAct, SIGNAL(triggered()), this, SIGNAL(movedRight()));

    mDeleteAct = new QAction(XdgIcon::fromTheme("dialog-close"), tr("Remove from quicklaunch"), this);
    connect(mDeleteAct, SIGNAL(triggered()), this, SLOT(selfRemove()));

    mMenu = new PopupMenu(mPlugin->panel()->compositorOpen(), this);
    mMenu->addAction(mAct);
    mMenu->addActions(mAct->addtitionalActions());
    mMenu->addSeparator();
    mMenu->addAction(mMoveLeftAct);
    mMenu->addAction(mMoveRightAct);
    mMenu->addSeparator();
    mMenu->addAction(mDeleteAct);
    mMenu->setWindowFlag(Qt::FramelessWindowHint);
    mMenu->setAttribute(Qt::WA_TranslucentBackground);

    // 阴影 软件启动栏右键添加圆角
    QGraphicsDropShadowEffect  * shadow_effect = new QGraphicsDropShadowEffect(this);
    shadow_effect->setOffset(0,0);
    QColor color1("#000000");
    color1.setAlphaF(0.2);
    shadow_effect->setColor(color1);
    shadow_effect->setBlurRadius(4);
    mMenu->setGraphicsEffect(shadow_effect);

    setContextMenuPolicy(Qt::CustomContextMenu);
    connect(this, SIGNAL(customContextMenuRequested(const QPoint&)),
            this, SLOT(this_customContextMenuRequested(const QPoint&)));
}


QuickLaunchButton::~QuickLaunchButton()
{
    //m_act->deleteLater();
}


QHash<QString,QString> QuickLaunchButton::settingsMap()
{
    Q_ASSERT(mAct);
    return mAct->settingsMap();
}

QMimeData *QuickLaunchButton::mimeData()
{
    QMimeData *mimedata = new QMimeData;
    QByteArray ba;
    QDataStream stream(&ba,QIODevice::WriteOnly);
    stream << this->settingsMap();
    mimedata->setData(mimeDataFormat(), ba);
    return mimedata;
}

void QuickLaunchButton::this_customContextMenuRequested(const QPoint & pos)
{
    this->setStyleSheet("QToolButton{"
                        "border: none;"
                        "margin: 2px;"
                        "margin-top: 0px;"
                        "margin-bottom: 0px;"
                        "padding: 4px;"
                        "background: rgba(255,255,255,0.12);}");

    mMoveLeftAct->setEnabled( mLaunch->indexOfButton(this) > 0);
    mMoveRightAct->setEnabled(mLaunch->indexOfButton(this) < mLaunch->countOfButtons() - 1);
    mMenu->setGeometry(mPlugin->panel()->calculatePopupWindowPos(mapToGlobal({0, 0}), mMenu->sizeHint()));
    mPlugin->willShowWindow(mMenu);
//    mMenu->popup(mPlugin->panel()->calculatePopupWindowPos(mapToGlobal({0, 0}), mMenu->sizeHint()).topLeft());

//    QString color = QString("%1, %2, %3, %4")
//            .arg(mPlugin->panel()->backgroundColor().red())
//            .arg(mPlugin->panel()->backgroundColor().green())
//            .arg(mPlugin->panel()->backgroundColor().blue())
//            .arg(1);
//    mMenu->setStyleSheet("background-color:rgba(" + color + "); border-radius: 4px;");

    mMenu->setCompositorOpen(mPlugin->panel()->compositorOpen());
    mMenu->exec();
    this->setStyleSheet("");
}

void QuickLaunchButton::selfRemove()
{
    Q_EMIT buttonDeleted();
}

void QuickLaunchButton::paintEvent(QPaintEvent *)
{
    // Do not paint that ugly "has menu" arrow
    QStylePainter p(this);
    QStyleOptionToolButton opt;
    initStyleOption(&opt);
    opt.features &= (~ QStyleOptionToolButton::HasMenu);
    p.drawComplexControl(QStyle::CC_ToolButton, opt);
}


void QuickLaunchButton::mousePressEvent(QMouseEvent *e)
{
    //重新处理LeftButton的press事件
    if (e->button() == Qt::LeftButton)
    {
        mDragStartPos = e->pos();
        //这里截断了press事件，需要手动设置一下qss样式
        this->setStyleSheet("QToolButton{"
                                "border: none;"
                                "margin: 2px;"
                                "margin-top: 0px;"
                                "margin-bottom: 0px;"
                                "padding: 4px;"
                                "background: rgba(255,255,255,0.12);}");
    }
    else {
        QToolButton::mousePressEvent(e);
    }
}


void QuickLaunchButton::mouseMoveEvent(QMouseEvent *e)
{
    mDragStart = false;
    if (!(e->buttons() & Qt::LeftButton))
    {
        return;
    }

    if ((e->pos() - mDragStartPos).manhattanLength() < QApplication::startDragDistance())
    {
        return;
    }

    mDragStart = true;

    QDrag *drag = new QDrag(this);
    drag->setMimeData(this->mimeData());
    drag->exec(Qt::MoveAction);

    //由于点击鼠标并开始拖拽后，不会触发btn的release事件，这里手动发送一下
    QMouseEvent event(QEvent::MouseButtonRelease, e->pos(),Qt::LeftButton,Qt::LeftButton,Qt::NoModifier);
    QApplication::sendEvent(this, &event);
}

void QuickLaunchButton::mouseReleaseEvent(QMouseEvent *e)
{
    //重新处理LeftButton的release事件
    if (e->button() == Qt::LeftButton)
    {
        if(mDragStart){
            mDragStart = false;
        }
        else {
            if(mAct->settingsMap().contains("desktop"))
            {
                //添加常用
                QDBusInterface * interface = new QDBusInterface(QString::fromLatin1("com.nde.panel.mainmenu"),
                                                             QString::fromLatin1("/com/nde/panel/mainmenu"),
                                                             QString::fromLatin1("com.nde.panel.mainmenu"),
                                                             QDBusConnection::sessionBus());
                if(interface->isValid()){
                    interface->call("addCommonAppCount", mAct->settingsMap().value("desktop"));
                }
            }
            mAct->execAction();
        }
        //取消press时设置的样式，执行qss中设置的样式
        this->setStyleSheet("");
    }
    else {
        QToolButton::mousePressEvent(e);
    }
}


void QuickLaunchButton::dragMoveEvent(QDragMoveEvent * e)
{
    if (e->mimeData() && e->mimeData()->hasFormat(mimeDataFormat()))
    {
        e->acceptProposedAction();
    }
    else
        e->ignore();
}

void QuickLaunchButton::dropEvent(QDropEvent *e)
{
    if (e->mimeData() && e->mimeData()->hasFormat(mimeDataFormat()))
    {
        e->acceptProposedAction();
    }
    else
        e->ignore();
}

void QuickLaunchButton::dragEnterEvent(QDragEnterEvent *e)
{
    if (e->mimeData() && e->mimeData()->hasFormat(mimeDataFormat()))
    {
        QuickLaunchButton* sourceBtn = qobject_cast<QuickLaunchButton*>(e->source());
        if(sourceBtn)
        {
            e->acceptProposedAction();
            Q_EMIT switchButtons(sourceBtn, this);
        }
        else
        {
            e->ignore();
        }
    }
    else
        e->ignore();
}
