/* BEGIN_COMMON_COPYRIGHT_HEADER
 * (c)LGPL2+
 *
 * Nde - a lightweight, Qt based, desktop toolset
 * https://lxqt.org
 *
 * Copyright: 2010-2011 Razor team
 * Authors:
 *   Petr Vanek <petr@scribus.info>
 *
 * This program or library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301 USA
 *
 * END_COMMON_COPYRIGHT_HEADER */

#include <QAction>
#include <QX11Info>
#include <nde-globalkeys.h>
#include <XdgIcon>
#include <Nde/Notification>
#include <KWindowSystem/KWindowSystem>
#include <KWindowSystem/NETWM>
#include "showdesktop.h"
#include "../panel/pluginsettings.h"
#include "QDebug"


#define DEFAULT_SHORTCUT "Control+Alt+D"
#define DEFAULT_SHORTCUT_WIN "Meta+D"
#define DEFAULT_SHORTCUT_MINIMIZE_WIN "Meta+M"

ShowDesktop::ShowDesktop(const INdePanelPluginStartupInfo &startupInfo) :
    QObject(),
    INdePanelPlugin(startupInfo)
{
    m_key = GlobalKeyShortcut::Client::instance()->addAction(QString(), QString("/panel/%1/show_hide").arg(settings()->group()), tr("Show desktop"), this);
    if (m_key)
    {
        connect(m_key, &GlobalKeyShortcut::Action::registrationFinished, this, &ShowDesktop::shortcutRegistered);
        connect(m_key, SIGNAL(activated()), this, SLOT(showHideAction()));
    }

    m_show_desktop = GlobalKeyShortcut::Client::instance()->addAction(QString(), QString("/panel/%1/show_desktop").arg(settings()->group()), tr("Show desktop"), this);
    if(m_show_desktop)
    {
        connect(m_show_desktop, &GlobalKeyShortcut::Action::registrationFinished, [this] {
            if(m_show_desktop->shortcut().isEmpty())
                m_show_desktop->changeShortcut(DEFAULT_SHORTCUT_WIN);
        });
        connect(m_show_desktop, SIGNAL(activated()), this, SLOT(showHideAction()));
    }

    // minimize all windows
    m_minimize_win = GlobalKeyShortcut::Client::instance()->addAction(QString(), QString("/panel/%1/minimize_win").arg(settings()->group()), tr("minimize all windows"), this);
    if(m_minimize_win)
    {
        connect(m_minimize_win, &GlobalKeyShortcut::Action::registrationFinished, [this] {
            if(m_minimize_win->shortcut().isEmpty())
                m_minimize_win->changeShortcut(DEFAULT_SHORTCUT_MINIMIZE_WIN);
        });
        connect(m_minimize_win, SIGNAL(activated()), this, SLOT(minimizeAllWindows()));
    }

    mTaskBarWidth = startupInfo.ndePanel->teskbarWidth();

    QAction * act = new QAction(this);
    connect(act, SIGNAL(triggered()), this, SLOT(showHideAction()));

    mButton = new ShowDesktopWidget();
    mButton->setDefaultAction(act);
    mButton->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);
    mButton->setAutoRaise(true);
    mButton->setProperty("direction", "horizontal");

}

void ShowDesktop::realign()
{
    if(this->panel()->position() > 1){
        mButton->setProperty("direction", "vertical");
        mButton->setFixedHeight(8);
        mButton->setFixedWidth(mTaskBarWidth);
    }else {
        mButton->setProperty("direction", "horizontal");
        mButton->setFixedWidth(8);
        mButton->setFixedHeight(mTaskBarWidth);
    }
    mButton->style()->unpolish(mButton);
    mButton->style()->polish(mButton);
}

void ShowDesktop::shortcutRegistered()
{
    if (m_key->shortcut().isEmpty())
    {
        m_key->changeShortcut(DEFAULT_SHORTCUT);
        if (m_key->shortcut().isEmpty())
        {
            Nde::Notification::notify(tr("Show Desktop: Global shortcut '%1' cannot be registered").arg(DEFAULT_SHORTCUT));
        }
    }
}

void ShowDesktop::toggleShowingDesktop()
{
    KWindowSystem::setShowingDesktop(!KWindowSystem::showingDesktop());
}

void ShowDesktop::minimizeAllWindows()
{
    const auto wnds = KWindowSystem::stackingOrder();
    for (auto const wnd: wnds)
    {
        KWindowSystem::minimizeWindow(wnd);
    }
}

void ShowDesktop::showHideAction()
{
    const auto wnds = KWindowSystem::stackingOrder();
    bool allIsMinimize = true;
    QMap<WId,bool> windowFlagListBak;
    for (auto const wnd: wnds){
        KWindowInfo info(wnd, NET::WMState | NET::XAWMState | NET::WMName);
        //Handling special windows...
        if(isWindowHighlightable(wnd)) {
        //if (info.valid() && info.name() != "nde-fileman" && info.name() != "Nde Panel"){
            if (!info.isMinimized()){
                allIsMinimize = false;
            }
            windowFlagListBak.insert(wnd, info.isMinimized());
        }
    }

    //存在窗口没最小化的做最小化处理，桌面空时再点击恢复上一次窗口状态
    if (!allIsMinimize) {
        minimizeAllWindows();
        windowFlagList.clear();
        windowFlagList = windowFlagListBak;
    }else {
        for (auto const wnd: wnds){
            if (windowFlagList.contains(wnd)){
                if (!windowFlagList.value(wnd)){
                    KWindowSystem::unminimizeWindow(wnd);
                }
            }
        }
    }
}

bool ShowDesktop::isWindowHighlightable(WId window)
{
    QFlags<NET::WindowTypeMask> ignoreList;
    ignoreList |= NET::DesktopMask;
    ignoreList |= NET::DockMask;
    ignoreList |= NET::SplashMask;
    ignoreList |= NET::ToolbarMask;
    ignoreList |= NET::MenuMask;
    ignoreList |= NET::PopupMenuMask;
    ignoreList |= NET::NotificationMask;

    KWindowInfo info(window, NET::WMWindowType | NET::WMState, NET::WM2TransientFor);
    if (!info.valid())
        return false;

    if (NET::typeMatchesMask(info.windowType(NET::AllTypesMask), ignoreList))
        return false;

    if (info.state() & NET::SkipTaskbar)
        return false;

    // WM_TRANSIENT_FOR hint not set - normal window
    WId transFor = info.transientFor();
    if (transFor == 0 || transFor == window || transFor == (WId) QX11Info::appRootWindow())
        return true;

    info = KWindowInfo(transFor, NET::WMWindowType);

    QFlags<NET::WindowTypeMask> normalFlag;
    normalFlag |= NET::NormalMask;
    normalFlag |= NET::DialogMask;
    normalFlag |= NET::UtilityMask;

    return !NET::typeMatchesMask(info.windowType(NET::AllTypesMask), normalFlag);
}
#undef DEFAULT_SHORTCUT
