/* BEGIN_COMMON_COPYRIGHT_HEADER
 * (c)LGPL2+
 *
 * Nde - a lightweight, Qt based, desktop toolset
 * https://lxqt.org
 *
 * Copyright: 2015 Nde team
 * Authors:
 *  Balázs Béla <balazsbela[at]gmail.com>
 *  Paulo Lieuthier <paulolieuthier@gmail.com>
 *
 * This program or library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301 USA
 *
 * END_COMMON_COPYRIGHT_HEADER */

#include "statusnotifierbutton.h"

#include <QDir>
#include <QFile>
#include <dbusmenu-qt5/dbusmenuimporter.h>
#include "../panel/indepanelplugin.h"
#include "../panel/shadowmanager.h"
#include "sniasync.h"
#include <XdgIcon>
#include <QDrag>
#include <QApplication>
#include <QStyle>

namespace
{
    /*! \brief specialized DBusMenuImporter to correctly create actions' icons based
     * on name
     */
    class MenuImporter : public DBusMenuImporter
    {
    public:
        using DBusMenuImporter::DBusMenuImporter;

    protected:
        virtual QIcon iconForName(const QString & name) override
        {
            return XdgIcon::fromTheme(name);
        }
    };
}

void setStatusNotifierButtonCompositer(QMenu * menu, bool compositer)
{
    if(!menu) return ;
    menu->setWindowFlag(Qt::FramelessWindowHint);
    menu->setAttribute(Qt::WA_TranslucentBackground);
    menu->setAttribute(Qt::WA_NoSystemBackground, false);
    ShadowManager::setShadow(menu);

    //右键菜单不支持透明效果
//    menu->setProperty("compositor", compositer);
//    menu->style()->unpolish(menu);
//    menu->style()->polish(menu);

    for(QAction * a : menu->actions())
    {
        if(a->menu()) {
            setStatusNotifierButtonCompositer(a->menu(), compositer);
        }
    }
}

bool StatusNotifierButton::sDraggging = false;
StatusNotifierButton::StatusNotifierButton(QString serviceAndPath, INdePanelPlugin* plugin, SniAsync *ifc, QWidget *parent)
    : QToolButton(parent),
    interface(ifc),
    mMenu(nullptr),
    mStatus(Passive),
    mFallbackIcon(QIcon::fromTheme("application-x-executable")),
    mPlugin(plugin),
    mServiceAndPath(serviceAndPath),
    isMoved(false)
{
    setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);
    setAutoRaise(true);
    setAcceptDrops(true);

    connect(interface, &SniAsync::NewIcon, this, &StatusNotifierButton::newIcon);
    connect(interface, &SniAsync::NewOverlayIcon, this, &StatusNotifierButton::newOverlayIcon);
    connect(interface, &SniAsync::NewAttentionIcon, this, &StatusNotifierButton::newAttentionIcon);
    connect(interface, &SniAsync::NewToolTip, this, &StatusNotifierButton::newToolTip);
    connect(interface, &SniAsync::NewStatus, this, &StatusNotifierButton::newStatus);

    interface->propertyGetAsync(QLatin1String("Menu"), [this] (QDBusObjectPath path) {
        if (!path.path().isEmpty())
        {
            MenuImporter *menuImport = new MenuImporter{interface->service(), path.path(), this};
            mMenu = menuImport->menu();
        }
    });

    interface->propertyGetAsync(QLatin1String("Status"), [this] (QString status) {
        newStatus(status);
    });

    interface->propertyGetAsync(QLatin1String("IconThemePath"), [this] (QString value) {
        mThemePath = value;
        //do the logic of icons after we've got the theme path
        refetchIcon(Active);
        refetchIcon(Passive);
        refetchIcon(NeedsAttention);
    });
    interface->propertyGetAsync(QLatin1String("Id"), [this] (QString id) {
        mTrayId = id;
    });
    interface->propertyGetAsync(QLatin1String("Title"), [this] (QString title) {
        mTitle = title;
    });
    newToolTip();
}

StatusNotifierButton::~StatusNotifierButton()
{
    delete interface;
}
QString StatusNotifierButton::trayId() const
{
    if(mTitle.isEmpty())
        return mTrayId;
    return mTitle;
}
void StatusNotifierButton::newIcon()
{
    //新版软件中心安装卸载软件时会一瞬间触发大量的输入法图标更新的信号，限制一下
    disconnect(interface, &SniAsync::NewIcon, this, &StatusNotifierButton::newIcon);
    refetchIcon(Passive);
}

void StatusNotifierButton::newOverlayIcon()
{
    refetchIcon(Active);
}

void StatusNotifierButton::newAttentionIcon()
{
    refetchIcon(NeedsAttention);
}

void StatusNotifierButton::refetchIcon(Status status)
{
    QString nameProperty, pixmapProperty;
    if (status == Active)
    {
        nameProperty = QLatin1String("OverlayIconName");
        pixmapProperty = QLatin1String("OverlayIconPixmap");
    }
    else if (status == NeedsAttention)
    {
        nameProperty = QLatin1String("AttentionIconName");
        pixmapProperty = QLatin1String("AttentionIconPixmap");
    }
    else // status == Passive
    {
        nameProperty = QLatin1String("IconName");
        pixmapProperty = QLatin1String("IconPixmap");
    }

    interface->propertyGetAsync(nameProperty, [this, status, pixmapProperty] (QString iconName) {
        QIcon nextIcon;
        if (!iconName.isEmpty())
        {
            if (QIcon::hasThemeIcon(iconName))
                nextIcon = QIcon::fromTheme(iconName);
            else
            {
                QDir themeDir(mThemePath);
                if (themeDir.exists())
                {
                    if (themeDir.exists(iconName + ".png"))
                        nextIcon.addFile(themeDir.filePath(iconName + ".png"));

                    if (themeDir.cd("hicolor") || (themeDir.cd("icons") && themeDir.cd("hicolor")))
                    {
                        const QStringList sizes = themeDir.entryList(QDir::AllDirs | QDir::NoDotAndDotDot);
                        for (const QString &dir : sizes)
                        {
                            const QStringList dirs = QDir(themeDir.filePath(dir)).entryList(QDir::AllDirs | QDir::NoDotAndDotDot);
                            for (const QString &innerDir : dirs)
                            {
                                QString file = themeDir.absolutePath() + "/" + dir + "/" + innerDir + "/" + iconName + ".png";
                                if (QFile::exists(file))
                                    nextIcon.addFile(file);
                            }
                        }
                    }
                }
            }

            switch (status)
            {
                case Active:
                    mOverlayIcon = nextIcon;
                    break;
                case NeedsAttention:
                    mAttentionIcon = nextIcon;
                    break;
                case Passive:
                    mIcon = nextIcon;
                    break;
            }

            resetIcon();
        }
        else
        {
            interface->propertyGetAsync(pixmapProperty, [this, status, pixmapProperty] (IconPixmapList iconPixmaps) {
                if (iconPixmaps.empty())
                    return;

                QIcon nextIcon;

                for (IconPixmap iconPixmap: iconPixmaps)
                {
                    if (!iconPixmap.bytes.isNull())
                    {
                        QImage image((uchar*) iconPixmap.bytes.data(), iconPixmap.width,
                                     iconPixmap.height, QImage::Format_ARGB32);

                        const uchar *end = image.constBits() + image.byteCount();
                        uchar *dest = reinterpret_cast<uchar*>(iconPixmap.bytes.data());
                        for (const uchar *src = image.constBits(); src < end; src += 4, dest += 4)
                            qToUnaligned(qToBigEndian<quint32>(qFromUnaligned<quint32>(src)), dest);
                        nextIcon.addPixmap(QPixmap::fromImage(image));
                    }
                }

                switch (status)
                {
                    case Active:
                        mOverlayIcon = nextIcon;
                        break;
                    case NeedsAttention:
                        mAttentionIcon = nextIcon;
                        break;
                    case Passive:
                        mIcon = nextIcon;
                        break;
                }

                resetIcon();
            });
        }
        QMetaObject::invokeMethod(this, "slotNewIconDone");
    });
}

void StatusNotifierButton::newToolTip()
{
    disconnect(interface, &SniAsync::NewToolTip, this, &StatusNotifierButton::newToolTip);
    interface->propertyGetAsync(QLatin1String("ToolTip"), [this] (ToolTip tooltip) {
        QString toolTipTitle = tooltip.title;
        if (!toolTipTitle.isEmpty())
            setToolTip(toolTipTitle);
        else
            interface->propertyGetAsync(QLatin1String("Title"), [this] (QString title) {
                // we should get here only in case the ToolTip.title was empty
                if (!title.isEmpty())
                    setToolTip(title);
            });
        QMetaObject::invokeMethod(this, "slotNewTooltipDone");
    });
}

void StatusNotifierButton::newStatus(QString status)
{
    Status newStatus;
    if (status == QLatin1String("Passive"))
        newStatus = Passive;
    else if (status == QLatin1String("Active"))
        newStatus = Active;
    else
        newStatus = NeedsAttention;

    if (mStatus == newStatus)
        return;

    mStatus = newStatus;
    resetIcon();
}

void StatusNotifierButton::slotNewIconDone()
{
    disconnect(interface, &SniAsync::NewIcon, this, &StatusNotifierButton::newIcon);
    connect(interface, &SniAsync::NewIcon, this, &StatusNotifierButton::newIcon);
}

void StatusNotifierButton::slotNewTooltipDone()
{
    disconnect(interface, &SniAsync::NewToolTip, this, &StatusNotifierButton::newToolTip);
    connect(interface, &SniAsync::NewToolTip, this, &StatusNotifierButton::newToolTip);
}
QMimeData * StatusNotifierButton::mimeData()
{
    QMimeData *mimedata = new QMimeData;
    QByteArray byteArray;
    QDataStream stream(&byteArray, QIODevice::WriteOnly);
    stream << mServiceAndPath;
    mimedata->setData(mimeDataFormat(), byteArray);
    return mimedata;
}
void StatusNotifierButton::dragEnterEvent(QDragEnterEvent *event)
{

    if (event->mimeData()->hasFormat(mimeDataFormat()))
    {
        Q_EMIT dragging(event->source(), event->pos());
        setAttribute(Qt::WA_UnderMouse, false);
        event->acceptProposedAction();
    }

    QToolButton::dragEnterEvent(event);
}
void StatusNotifierButton::dragMoveEvent(QDragMoveEvent *event)
{
    if (event->mimeData()->hasFormat(mimeDataFormat()))
    {
        Q_EMIT dragging(event->source(), event->pos());
        setAttribute(Qt::WA_UnderMouse, false);
    }
    QToolButton::dragMoveEvent(event);
}
void StatusNotifierButton::dragLeaveEvent(QDragLeaveEvent *event)
{
     QToolButton::dragLeaveEvent(event);
}
void StatusNotifierButton::dropEvent(QDropEvent *event)
{
    if (event->mimeData()->hasFormat(mimeDataFormat()))
    {
        QString  serviceAndPath;
        QDataStream stream(event->mimeData()->data(mimeDataFormat()));
        stream >> serviceAndPath;
        Q_EMIT dropped(event->source(), event->pos(), serviceAndPath);
        setAttribute(Qt::WA_UnderMouse, false);
    }
    QToolButton::dropEvent(event);
}
void StatusNotifierButton::mouseMoveEvent(QMouseEvent *event)
{
    if (!(event->buttons() & Qt::LeftButton))
        return;

    if ((event->pos() - mDragStartPosition).manhattanLength() < QApplication::startDragDistance())
        return;

    QDrag *drag = new QDrag(this);
    drag->setMimeData(mimeData());
    QIcon ico = icon();
    QPixmap img = ico.pixmap(ico.actualSize({18, 18}));
    drag->setPixmap(img);
    switch (mPlugin->panel()->position())
    {
        case INdePanel::PositionLeft:
        case INdePanel::PositionTop:
            drag->setHotSpot({0, 0});
            break;
        case INdePanel::PositionRight:
        case INdePanel::PositionBottom:
            drag->setHotSpot(img.rect().bottomRight());
            break;
    }

    sDraggging = true;
    isMoved = true;
    drag->exec();
    QMouseEvent mouseEvent(QEvent::MouseButtonRelease, event->pos(), Qt::LeftButton, Qt::LeftButton, Qt::NoModifier);
    QApplication::sendEvent(this, &mouseEvent);
    drag->deleteLater();
    sDraggging = false;

    QAbstractButton::mouseMoveEvent(event);
}
void StatusNotifierButton::contextMenuEvent(QContextMenuEvent* event)
{
    //XXX: avoid showing of parent's context menu, we are (optionaly) providing context menu on mouseReleaseEvent
    //QWidget::contextMenuEvent(event);
}
void StatusNotifierButton::mousePressEvent(QMouseEvent *event)
{
    const Qt::MouseButton b = event->button();

    if (Qt::LeftButton == b)
        mDragStartPosition = event->pos();

    QToolButton::mousePressEvent(event);
}
void StatusNotifierButton::mouseReleaseEvent(QMouseEvent *event)
{
    if(isMoved)
    {
        isMoved = false;
        QToolButton::mouseReleaseEvent(event);
        return;
    }
    if (event->button() == Qt::LeftButton)
    {
        Q_EMIT activated();
        interface->Activate(QCursor::pos().x(), QCursor::pos().y());
    }
    else if (event->button() == Qt::MidButton)
    {
        interface->SecondaryActivate(QCursor::pos().x(), QCursor::pos().y());
    }
    else if (Qt::RightButton == event->button())
    {
        if (mMenu)
        {
            setStatusNotifierButtonCompositer(mMenu, mPlugin->panel()->compositorOpen());
            mPlugin->willShowWindow(mMenu);
            mMenu->popup(mPlugin->panel()->calculatePopupWindowPos(QCursor::pos(), mMenu->sizeHint()).topLeft());
        }
        else
        {
            interface->ContextMenu(QCursor::pos().x(), QCursor::pos().y());
        }
    }

    QToolButton::mouseReleaseEvent(event);
}

void StatusNotifierButton::wheelEvent(QWheelEvent *event)
{
    interface->Scroll(event->delta(), "vertical");
}

void StatusNotifierButton::resetIcon()
{
    if (mStatus == Active && !mOverlayIcon.isNull())
        setIcon(mOverlayIcon);
    else if (mStatus == NeedsAttention && !mAttentionIcon.isNull())
        setIcon(mAttentionIcon);
    else if (!mIcon.isNull()) // mStatus == Passive
        setIcon(mIcon);
    else if (!mOverlayIcon.isNull())
        setIcon(mOverlayIcon);
    else if (!mAttentionIcon.isNull())
        setIcon(mAttentionIcon);
    else
        setIcon(QIcon());
}
