/* BEGIN_COMMON_COPYRIGHT_HEADER
 * (c)LGPL2+
 *
 * Nde - a lightweight, Qt based, desktop toolset
 * https://lxqt.org
 *
 * Copyright: 2015 Nde team
 * Authors:
 *  Balázs Béla <balazsbela[at]gmail.com>
 *  Paulo Lieuthier <paulolieuthier@gmail.com>
 *
 * This program or library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301 USA
 *
 * END_COMMON_COPYRIGHT_HEADER */

#include "statusnotifierwidget.h"
#include "../panel/pluginsettings.h"
#include <QApplication>
#include <QDebug>
#include "../panel/indepanelplugin.h"
#include <QMimeData>

StatusNotifierWidget::StatusNotifierWidget(INdePanelPlugin *plugin, QString name, QWidget *parent) :
    QWidget(parent),
    mPlugin(plugin),
    mName(name)
{
    setAcceptDrops(true);
    mLayout = new Nde::GridLayout(this);
    setLayout(mLayout);
}

StatusNotifierWidget::~StatusNotifierWidget()
{

}

void StatusNotifierWidget::itemAdded(QString serviceAndPath, SniAsync *interface, int index)
{
    StatusNotifierButton *button = new StatusNotifierButton(serviceAndPath, mPlugin, interface, this);
    connect(button, &StatusNotifierButton::dragging, this, [=] (QObject * dragSource, QPoint const & pos) {
        buttonMove(qobject_cast<StatusNotifierButton *>(sender()), qobject_cast<StatusNotifierButton *>(dragSource), pos);
    });
    connect(button, &StatusNotifierButton::dropped, this, [=] (QObject * dragSource, QPoint const & pos, QString & serviceAndPath) {
        buttonDropped(qobject_cast<StatusNotifierButton *>(sender()), qobject_cast<StatusNotifierButton *>(dragSource), pos, serviceAndPath);
    });
    connect(button, &StatusNotifierButton::activated, this, &StatusNotifierWidget::itemActivated);
    layoutAdded(serviceAndPath, button, index);
}
void StatusNotifierWidget::layoutAdded(const QString & serviceAndPath, StatusNotifierButton *button, int index)
{
    mServices.insert(serviceAndPath, button);
    if(index > 0)
    {
        if(index > mLayout->count())
            index = mLayout->count();
        mLayout->addWidget(button);
        mLayout->moveItem(mLayout->indexOf(button), index, false);
    }
    else
    {
        mLayout->addWidget(button);
    }
    button->show();

}
int StatusNotifierWidget::index(const QString & serviceAndPath) const
{
    StatusNotifierButton *button = mServices.value(serviceAndPath);
    return mLayout->indexOf(button);
}
QString StatusNotifierWidget::trayId(QString service) const
{
    StatusNotifierButton *button = mServices.value(service);
    if(button)
        return button->trayId();
    return QString();
}

void StatusNotifierWidget::saveOrder()
{
    QStringList orders;
    for(int i = 0; i < mLayout->count(); i++)
    {
        StatusNotifierButton * btn = qobject_cast<StatusNotifierButton*>(mLayout->itemAt(i)->widget());
        if(btn)
        {
            orders.append(btn->trayId());
        }
    }
    mPlugin->settings()->setValue(this->orderName(), orders);
}

void  StatusNotifierWidget::buttonDropped(StatusNotifierButton * dst, StatusNotifierButton * src, QPoint const & pos, QString &serviceAndPath)
{
    int src_index;
    if (-1 != (src_index = mLayout->indexOf(src)) && mServices.contains(serviceAndPath))
        return;
    int dst_index = mLayout->indexOf(dst);
    Q_EMIT dropButton(serviceAndPath, dst_index);
}
void StatusNotifierWidget::buttonMove(StatusNotifierButton * dst, StatusNotifierButton * src, QPoint const & pos)
{
    int src_index;
    if (!src || -1 == (src_index = mLayout->indexOf(src)))
    {
        qDebug() << "Dropped invalid";
        return;
    }

    const int size = mLayout->count();
    Q_ASSERT(0 < size);
    int dst_index = mLayout->indexOf(dst);
    if(-1 == dst_index)
        return;
    if (src_index < dst_index)
    {
        if (size == dst_index
                || src_index + 1 != dst_index)
        {
            --dst_index;
        } else
        {
            //switching positions of next standing
            const int tmp_index = src_index;
            src_index = dst_index;
            dst_index = tmp_index;
        }
    }

    if (dst_index == src_index
            || mLayout->animatedMoveInProgress()
       )
        return;
    mLayout->moveItem(src_index, dst_index, true);

    saveOrder();
}
void StatusNotifierWidget::itemRemoved(const QString &serviceAndPath)
{
    StatusNotifierButton *button = mServices.take(serviceAndPath);
    if (button)
    {
        button->deleteLater();
        layout()->removeWidget(button);
    }
}

StatusNotifierButton *StatusNotifierWidget::layoutRemoved(const QString & serviceAndPath)
{
    StatusNotifierButton *btn = mServices.take(serviceAndPath);
    if(btn)
    {
        layout()->removeWidget(btn);
    }
    return btn;
}
void StatusNotifierWidget::dropEvent(QDropEvent *event)
{
    QString  serviceAndPath;
    QDataStream stream(event->mimeData()->data(mimeDataFormat()));
    stream >> serviceAndPath;
    if(mServices.contains(serviceAndPath))
    {
        StatusNotifierButton *button = mServices.value(serviceAndPath, nullptr);
        if (button && event->mimeData()->hasFormat(mimeDataFormat()))
        {
            Q_EMIT button->dragging(event->source(), event->pos());
            setAttribute(Qt::WA_UnderMouse, false);
        }
    }
    else
    {
        Q_EMIT dropButton(serviceAndPath);
    }
}
void StatusNotifierWidget::arrowDropped(QString & serviceAndPath)
{
    if(mServices.size() <= 0)
    {
        Q_EMIT dropButton(serviceAndPath);
    }
}
void StatusNotifierWidget::dragEnterEvent(QDragEnterEvent *event)
{

    if (event->mimeData()->hasFormat(mimeDataFormat()))
    {
        event->acceptProposedAction();
    }
}
void StatusNotifierWidget::dragLeaveEvent(QDragLeaveEvent *event)
{
}


void StatusNotifierWidget::itemActivated()
{
    //do nothing
}
