/* BEGIN_COMMON_COPYRIGHT_HEADER
 * (c)LGPL2+
 *
 * Nde - a lightweight, Qt based, desktop toolset
 * https://lxqt.org
 *
 * Copyright: 2011 Razor team
 *            2014 Nde team
 * Authors:
 *   Alexander Sokoloff <sokoloff.a@gmail.com>
 *   Kuzma Shapran <kuzma.shapran@gmail.com>
 *
 * This program or library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301 USA
 *
 * END_COMMON_COPYRIGHT_HEADER */

#include "ndetaskbutton.h"
#include "ndetaskgroup.h"
#include "ndetaskbar.h"

#include <Nde/Settings>

#include <QDebug>
#include <XdgIcon>
#include <QTimer>
#include <QMenu>
#include <QAction>
#include <QContextMenuEvent>
#include <QProcess>
#include <QPainter>
#include <QDrag>
#include <QMouseEvent>
#include <QApplication>
#include <QDragEnterEvent>
#include <QStylePainter>
#include <QStyleOptionToolButton>
#include <QDesktopWidget>

#include "ndetaskbutton.h"
#include "ndetaskgroup.h"
#include "ndetaskbar.h"
#include "../panel/popupmenu.h"
#include <KWindowSystem/KWindowSystem>
// Necessary for closeApplication()
#include <KWindowSystem/NETWM>
#include <QX11Info>

bool NdeTaskButton::sDraggging = false;

/************************************************

************************************************/
void LeftAlignedTextStyle::drawItemText(QPainter * painter, const QRect & rect, int flags
            , const QPalette & pal, bool enabled, const QString & text
            , QPalette::ColorRole textRole) const
{
    QString txt = QFontMetrics(painter->font()).elidedText(text, Qt::ElideRight, rect.width());
    return QProxyStyle::drawItemText(painter, rect, (flags & ~Qt::AlignHCenter) | Qt::AlignLeft, pal, enabled, txt, textRole);
}


/************************************************

************************************************/
NdeTaskButton::NdeTaskButton(const WId window, NdeTaskBar * taskbar, bool autoRotation, QWidget *parent) :
    QToolButton(parent),
    mPreventPopup(false),
    mWindow(window),
    mAutoRotation(autoRotation),
    mUrgencyHint(false),
    mOrigin(Qt::TopLeftCorner),
    mDrawPixmap(false),
    mParentTaskBar(taskbar),
    mPlugin(mParentTaskBar->plugin()),
    currentPosition(-1),
    currentPanelPage(0),
    mDNDTimer(new QTimer(this))
//    mButtonStatus(BUTTON_UNFIXED)
{
    Q_ASSERT(taskbar);
    updateText();
    updateIcon();
    initButton();
}
NdeTaskButton::NdeTaskButton(const XdgDesktopFile &xdg, NdeTaskBar * taskbar, bool autoRotation, QWidget *parent):
    QToolButton(parent),
    mWindow(0),
    mAutoRotation(autoRotation),
    mUrgencyHint(false),
    mOrigin(Qt::TopLeftCorner),
    mDrawPixmap(false),
    mParentTaskBar(taskbar),
    mPlugin(mParentTaskBar->plugin()),
    currentPosition(-1),
    currentPanelPage(0),
    mDNDTimer(new QTimer(this)),
    mXdg(xdg)
//    mButtonStatus(BUTTON_FIXED)
{
    QString title(xdg.localizedValue("Name").toString());
    QString gn(xdg.localizedValue("GenericName").toString());
    if (!gn.isEmpty())
        title += " (" + gn + ")";
    setText(title);
    setIcon(xdg.icon(XdgIcon::defaultApplicationIcon()));

    initButton();
}
void NdeTaskButton::initButton()
{
    setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);
    setMinimumWidth(1);
    setMinimumHeight(1);
    setToolButtonStyle(Qt::ToolButtonTextBesideIcon);
    setAcceptDrops(true);
    setCheckable(true);

    this->setProperty("autorotation", mAutoRotation);
    updateButtonStyle(mAutoRotation, parentTaskBar()->panel()->position());

    isMoved = false;
    mDNDTimer->setSingleShot(true);
    mDNDTimer->setInterval(700);
    connect(mDNDTimer, SIGNAL(timeout()), this, SLOT(activateWithDraggable()));
    connect(Nde::Settings::globalSettings(), SIGNAL(iconThemeChanged()), this, SLOT(updateIcon()));
    connect(mParentTaskBar, &NdeTaskBar::iconByClassChanged, this, &NdeTaskButton::updateIcon);

    m_mouseLeaseTimer = new QTimer(this);
    m_mouseLeaseTimer->setSingleShot(true);
    m_mouseLeaseTimer->setInterval(200);
}

/************************************************

************************************************/
NdeTaskButton::~NdeTaskButton()
{
}

/************************************************

 ************************************************/
void NdeTaskButton::updateText()
{
    KWindowInfo info(mWindow, NET::WMVisibleName | NET::WMName);
    QString title = info.visibleName().isEmpty() ? info.name() : info.visibleName();
    setText(title.replace("&", "&&"));
    setToolTip(title);
}

/************************************************

 ************************************************/
void NdeTaskButton::updateIcon()
{
    QIcon ico;
    if (mParentTaskBar->isIconByClass())
    {
        ico = XdgIcon::fromTheme(QString::fromUtf8(KWindowInfo{mWindow, 0, NET::WM2WindowClass}.windowClassClass()).toLower());
    }
    if (ico.isNull())
    {
	if (name() == "nde-fileman"){
		ico = QIcon::fromTheme("folder");
	}else if (name() == "ndeterminal"){
		ico = QIcon::fromTheme("utilities-terminal");
	}else{
        	ico = KWindowSystem::icon(mWindow);
	}
    }

    setIcon(ico.isNull() ? XdgIcon::defaultApplicationIcon() : ico);
}

/************************************************

 ************************************************/
void NdeTaskButton::refreshIconGeometry(QRect const & geom)
{
    NETWinInfo info(QX11Info::connection(),
                    windowId(),
                    (WId) QX11Info::appRootWindow(),
                    NET::WMIconGeometry,
                    0);
    NETRect const curr = info.iconGeometry();
    if (curr.pos.x != geom.x() || curr.pos.y != geom.y()
            || curr.size.width != geom.width() || curr.size.height != geom.height())
    {
        NETRect nrect;
        nrect.pos.x = geom.x();
        nrect.pos.y = geom.y();
        nrect.size.height = geom.height();
        nrect.size.width = geom.width();
        info.setIconGeometry(nrect);
    }
}

/************************************************

 ************************************************/
void NdeTaskButton::dragEnterEvent(QDragEnterEvent *event)
{
    // It must be here otherwise dragLeaveEvent and dragMoveEvent won't be called
    // on the other hand drop and dragmove events of parent widget won't be called
    event->acceptProposedAction();
    if (event->mimeData() && event->mimeData()->hasFormat(mimeDataFormat()))
    {
        Q_EMIT dragging(event->source(), event->pos());
        setAttribute(Qt::WA_UnderMouse, false);
    }
    else
    {
        mDNDTimer->start();
    }

    QToolButton::dragEnterEvent(event);
}

void NdeTaskButton::dragMoveEvent(QDragMoveEvent * event)
{
    event->acceptProposedAction();
    if (event->mimeData() && event->mimeData()->hasFormat(mimeDataFormat()))
    {
        Q_EMIT dragging(event->source(), event->pos());
        setAttribute(Qt::WA_UnderMouse, false);
    }
}

void NdeTaskButton::dragLeaveEvent(QDragLeaveEvent *event)
{
    mDNDTimer->stop();
    QToolButton::dragLeaveEvent(event);
}

void NdeTaskButton::dropEvent(QDropEvent *event)
{
    mDNDTimer->stop();
    if (event->mimeData() && event->mimeData()->hasFormat(mimeDataFormat()))
    {
        setAttribute(Qt::WA_UnderMouse, false);
    }
    QToolButton::dropEvent(event);
}

/************************************************

 ************************************************/
void NdeTaskButton::mousePressEvent(QMouseEvent* event)
{
    const Qt::MouseButton b = event->button();

    if (Qt::LeftButton == b)
        mDragStartPosition = event->pos();
    else if (Qt::MidButton == b && parentTaskBar()->closeOnMiddleClick())
        closeApplication();
//    QToolButton::mousePressEvent(event);
}

/************************************************

 ************************************************/
void NdeTaskButton::mouseReleaseEvent(QMouseEvent* event)
{
    if (!sDraggging && event->button() == Qt::LeftButton && !m_mouseLeaseTimer->isActive())
    {
//        if(mButtonStatus == BUTTON_FIXED)
//        {
//            qint64 pid =  mXdg.startDetached();
//            if(!pid)
//            {
//                qWarning() << mXdg.fileName() << "start failed.";
//            }
//            mParentTaskBar->addProcessPid(mXdg.expandExecString().at(0), pid);
//        }
        if (isChecked())
            minimizeApplication();
        else
            raiseApplication();
        m_mouseLeaseTimer->start();
    }
    QToolButton::mouseReleaseEvent(event);
}

/************************************************

 ************************************************/
QMimeData * NdeTaskButton::mimeData()
{
    QMimeData *mimedata = new QMimeData;
    QByteArray ba;
    QDataStream stream(&ba,QIODevice::WriteOnly);
    stream << (qlonglong)(mWindow);
    mimedata->setData(mimeDataFormat(), ba);
    return mimedata;
}
QString NdeTaskButton::name() const
{
    KWindowInfo info(mWindow, NET::WMVisibleName | NET::WMName);
    KWindowInfo info1(mWindow, nullptr, NET::WM2WindowClass);
    QString title = info.visibleName().isEmpty() ? info.name() : info.visibleName();
    return mXdg.localizedValue("Name").toString().isEmpty() ?  info1.windowClassClass() : mXdg.localizedValue("Name").toString();
}
/************************************************

 ************************************************/
void NdeTaskButton::mouseMoveEvent(QMouseEvent* event)
{
    if (!(event->buttons() & Qt::LeftButton))
        return;

    if ((event->pos() - mDragStartPosition).manhattanLength() < QApplication::startDragDistance())
        return;

    QDrag *drag = new QDrag(this);
    drag->setMimeData(mimeData());
    QIcon ico = icon();
    QPixmap img = ico.pixmap(ico.actualSize({32, 32}));
    drag->setPixmap(img);
    switch (parentTaskBar()->panel()->position())
    {
        case INdePanel::PositionLeft:
        case INdePanel::PositionTop:
            drag->setHotSpot({0, 0});
            break;
        case INdePanel::PositionRight:
        case INdePanel::PositionBottom:
            drag->setHotSpot(img.rect().bottomRight());
            break;
    }

    sDraggging = true;
    drag->exec(Qt::MoveAction);
    isMoved = true;

    // if button is dropped out of panel (e.g. on desktop)
    // it is not deleted automatically by Qt
    drag->deleteLater();
    sDraggging = false;

    QAbstractButton::mouseMoveEvent(event);

    //由于点击鼠标并开始拖拽后，不会触发btn的release事件，这里手动发送一下
    QMouseEvent e(QEvent::MouseButtonRelease, event->pos(),Qt::LeftButton,Qt::LeftButton,Qt::NoModifier);
    QApplication::sendEvent(this, &e);
}

/************************************************

 ************************************************/
bool NdeTaskButton::isApplicationHidden() const
{
    KWindowInfo info(mWindow, NET::WMState);
    return (info.state() & NET::Hidden);
}

/************************************************

 ************************************************/
bool NdeTaskButton::isApplicationActive() const
{
    return KWindowSystem::activeWindow() == mWindow;
}

/************************************************

 ************************************************/
void NdeTaskButton::activateWithDraggable()
{
    // raise app in any time when there is a drag
    // in progress to allow drop it into an app
    raiseApplication();
    KWindowSystem::forceActiveWindow(mWindow);
}

/************************************************

 ************************************************/
void NdeTaskButton::raiseApplication()
{
    KWindowInfo info(mWindow, NET::WMDesktop | NET::WMState | NET::XAWMState);
    if (parentTaskBar()->raiseOnCurrentDesktop() && info.isMinimized())
    {
        KWindowSystem::setOnDesktop(mWindow, KWindowSystem::currentDesktop());
    }
    else
    {
        int winDesktop = info.desktop();
        if (KWindowSystem::currentDesktop() != winDesktop)
            KWindowSystem::setCurrentDesktop(winDesktop);
    }
    KWindowSystem::forceActiveWindow(mWindow);

    setUrgencyHint(false);
}

/************************************************

 ************************************************/
void NdeTaskButton::minimizeApplication()
{
    KWindowSystem::minimizeWindow(mWindow);
}

/************************************************

 ************************************************/
void NdeTaskButton::maximizeApplication()
{
    QAction* act = qobject_cast<QAction*>(sender());
    if (!act)
        return;

    int state = act->data().toInt();
    switch (state)
    {
        case NET::MaxHoriz:
            KWindowSystem::setState(mWindow, NET::MaxHoriz);
            break;

        case NET::MaxVert:
            KWindowSystem::setState(mWindow, NET::MaxVert);
            break;

        default:
            KWindowSystem::setState(mWindow, NET::Max);
            break;
    }

    if (!isApplicationActive())
        raiseApplication();
}

/************************************************

 ************************************************/
void NdeTaskButton::deMaximizeApplication()
{
    KWindowSystem::clearState(mWindow, NET::Max);

    if (!isApplicationActive())
        raiseApplication();
}

/************************************************

 ************************************************/
void NdeTaskButton::shadeApplication()
{
    KWindowSystem::setState(mWindow, NET::Shaded);
}

/************************************************

 ************************************************/
void NdeTaskButton::unShadeApplication()
{
    KWindowSystem::clearState(mWindow, NET::Shaded);
}

/************************************************

 ************************************************/
void NdeTaskButton::closeApplication()
{
    // FIXME: Why there is no such thing in KWindowSystem??
    NETRootInfo(QX11Info::connection(), NET::CloseWindow).closeWindowRequest(mWindow);
}

/************************************************

 ************************************************/
void NdeTaskButton::setApplicationLayer()
{
    QAction* act = qobject_cast<QAction*>(sender());
    if (!act)
        return;

    int layer = act->data().toInt();
    switch(layer)
    {
        case NET::KeepAbove:
            KWindowSystem::clearState(mWindow, NET::KeepBelow);
            KWindowSystem::setState(mWindow, NET::KeepAbove);
            break;

        case NET::KeepBelow:
            KWindowSystem::clearState(mWindow, NET::KeepAbove);
            KWindowSystem::setState(mWindow, NET::KeepBelow);
            break;

        default:
            KWindowSystem::clearState(mWindow, NET::KeepBelow);
            KWindowSystem::clearState(mWindow, NET::KeepAbove);
            break;
    }
}

/************************************************

 ************************************************/
void NdeTaskButton::moveApplicationToDesktop()
{
    QAction* act = qobject_cast<QAction*>(sender());
    if (!act)
        return;

    bool ok;
    int desk = act->data().toInt(&ok);

    if (!ok)
        return;

    KWindowSystem::setOnDesktop(mWindow, desk);
}

/************************************************

 ************************************************/
void NdeTaskButton::moveApplication()
{
    KWindowInfo info(mWindow, NET::WMDesktop);
    if (!info.isOnCurrentDesktop())
        KWindowSystem::setCurrentDesktop(info.desktop());
    if (isMinimized())
        KWindowSystem::unminimizeWindow(mWindow);
    KWindowSystem::forceActiveWindow(mWindow);
    const QRect& g = KWindowInfo(mWindow, NET::WMGeometry).geometry();
    int X = g.center().x();
    int Y = g.center().y();
    QCursor::setPos(X, Y);
    NETRootInfo(QX11Info::connection(), NET::WMMoveResize).moveResizeRequest(mWindow, X, Y, NET::Move);
}

/************************************************

 ************************************************/
void NdeTaskButton::resizeApplication()
{
    KWindowInfo info(mWindow, NET::WMDesktop);
    if (!info.isOnCurrentDesktop())
        KWindowSystem::setCurrentDesktop(info.desktop());
    if (isMinimized())
        KWindowSystem::unminimizeWindow(mWindow);
    KWindowSystem::forceActiveWindow(mWindow);
    const QRect& g = KWindowInfo(mWindow, NET::WMGeometry).geometry();
    int X = g.bottomRight().x();
    int Y = g.bottomRight().y();
    QCursor::setPos(X, Y);
    NETRootInfo(QX11Info::connection(), NET::WMMoveResize).moveResizeRequest(mWindow, X, Y, NET::BottomRight);
}

/************************************************

 ************************************************/
void NdeTaskButton::contextMenuEvent(QContextMenuEvent* event)
{
    if (event->modifiers().testFlag(Qt::ControlModifier))
    {
        event->ignore();
        return;
    }

    QMenu * menu = new PopupMenu(plugin()->panel()->compositorOpen(), this);
    menu->setAttribute(Qt::WA_DeleteOnClose);

//    if(isFixed())
//    {
//        QAction *a = menu->addAction(tr("Unpin To TaskBar"));
//        connect(a, SIGNAL(triggered()), this, SLOT(unpinApp()));
//    }
//    if(buttonStatus() != BUTTON_FIXED)
//    {
//        QAction *a = menu->addAction(tr("&Close"));
//        connect(a, SIGNAL(triggered()), this, SLOT(closeApplication()));
//    }
    QAction *a = menu->addAction(tr("&Close"));
    connect(a, SIGNAL(triggered()), this, SLOT(closeApplication()));
    connect(menu, &QMenu::aboutToHide, [this] {
        mPreventPopup = false;
        update();
    });
    menu->setGeometry(mParentTaskBar->panel()->calculatePopupWindowPos(mapToGlobal(event->pos()), menu->sizeHint()));
    mPlugin->willShowWindow(menu);
    menu->exec();
}
//void NdeTaskButton::unpinApp()
//{
//    QStringList fixedApps = mPlugin->settings()->value("FixedApps").toStringList();
//    QStringList desktopFile = mXdg.fileName().split('/');
//    QString desktopName = desktopFile.at(desktopFile.size() - 1);
//    for(auto iter = fixedApps.begin(); iter != fixedApps.end();)
//    {
//        QString desktop = *iter;
//        if(desktop.contains(desktopName))
//        {
//            fixedApps.erase(iter++);
//            mPlugin->settings()->setValue("FixedApps", fixedApps);
//            break;
//        }else
//            iter++;
//    }

//    mParentTaskBar->removeGroupButton(mXdg.expandExecString().at(0));
//}
//bool NdeTaskButton::isFixed() const
//{
//    QStringList fixedApps = mPlugin->settings()->value("FixedApps").toStringList();
//    if(mXdg.fileName().isEmpty())
//        return false;
//    QStringList desktopFile = mXdg.fileName().split('/');
//    QString desktopName = desktopFile.at(desktopFile.size() - 1);
//    for(auto iter = fixedApps.begin(); iter != fixedApps.end(); iter++)
//    {
//        QString desktop = *iter;
//        if( desktop.contains(desktopName))
//        {
//            return true;
//        }
//    }
//    return false;
//}
/************************************************

 ************************************************/
void NdeTaskButton::setUrgencyHint(bool set)
{
    if (mUrgencyHint == set)
        return;

    if (!set)
        KWindowSystem::demandAttention(mWindow, false);

    mUrgencyHint = set;
    setProperty("urgent", set);
    style()->unpolish(this);
    style()->polish(this);
    update();
}

/************************************************

 ************************************************/
bool NdeTaskButton::isOnDesktop(int desktop) const
{
    return KWindowInfo(mWindow, NET::WMDesktop).isOnDesktop(desktop);
}

bool NdeTaskButton::isOnCurrentScreen() const
{
    return QApplication::desktop()->screenGeometry(parentTaskBar()).intersects(KWindowInfo(mWindow, NET::WMFrameExtents).frameGeometry());
}

bool NdeTaskButton::isMinimized() const
{
    return KWindowInfo(mWindow,NET::WMState | NET::XAWMState).isMinimized();
}

Qt::Corner NdeTaskButton::origin() const
{
    return mOrigin;
}

void NdeTaskButton::setOrigin(Qt::Corner newOrigin)
{
    if (mOrigin != newOrigin)
    {
        mOrigin = newOrigin;
        update();
    }
}
void NdeTaskButton::updateButtonStyle(bool autoRotation, int position)
{
//    if(currentPosition == position || mButtonStatus == BUTTON_FIXED)
    bool needPolish = false;
    if(currentPosition != position)
    {
        needPolish = true;
        currentPosition = position;
        switch (position)
        {
        case INdePanel::PositionTop:
            this->setProperty("direction", "top");
            break;
        case INdePanel::PositionLeft:
            this->setProperty("direction", "left");
            break;
        case INdePanel::PositionRight:
            this->setProperty("direction", "right");
            break;
        case INdePanel::PositionBottom:
            this->setProperty("direction", "bottom");
            break;
        default:
            this->setProperty("direction", "");
            break;
        }
    }

    if(mAutoRotation != autoRotation)
    {
        needPolish = true;
        mAutoRotation = autoRotation;
        this->setProperty("autorotation", mAutoRotation);
    }

    if(needPolish)
    {
        this->style()->unpolish(this);
        this->style()->polish(this);
    }
}
void NdeTaskButton::setAutoRotation(bool autoRotation, INdePanel::Position position)
{
    updateButtonStyle(autoRotation, position);
    if (autoRotation)
    {
        switch (position)
        {
        case INdePanel::PositionTop:
        case INdePanel::PositionBottom:
            setOrigin(Qt::TopLeftCorner);
            break;

        case INdePanel::PositionLeft:
            setOrigin(Qt::BottomLeftCorner);
            break;

        case INdePanel::PositionRight:
            setOrigin(Qt::TopRightCorner);
            break;
        }
    }
    else
        setOrigin(Qt::TopLeftCorner);
}

void NdeTaskButton::paintEvent(QPaintEvent *event)
{
    if (mOrigin == Qt::TopLeftCorner)
    {
        QToolButton::paintEvent(event);
        return;
    }

    QSize sz = size();
    QSize adjSz = sz;
    QTransform transform;
    QPoint originPoint;

    switch (mOrigin)
    {
    case Qt::TopLeftCorner:
        transform.rotate(0.0);
        originPoint = QPoint(0.0, 0.0);
        break;

    case Qt::TopRightCorner:
        transform.rotate(90.0);
        originPoint = QPoint(0.0, -sz.width());
        adjSz.transpose();
        break;

    case Qt::BottomRightCorner:
        transform.rotate(180.0);
        originPoint = QPoint(-sz.width(), -sz.height());
        break;

    case Qt::BottomLeftCorner:
        transform.rotate(270.0);
        originPoint = QPoint(-sz.height(), 0.0);
        adjSz.transpose();
        break;
    }

    bool drawPixmapNextTime = false;

    if (!mDrawPixmap)
    {
        mPixmap = QPixmap(adjSz);
        mPixmap.fill(QColor(0, 0, 0, 0));

        if (adjSz != sz)
            resize(adjSz); // this causes paint event to be repeated - next time we'll paint the pixmap to the widget surface.

        // copied from QToolButton::paintEvent   {
        QStylePainter painter(&mPixmap, this);
        QStyleOptionToolButton opt;
        initStyleOption(&opt);
        painter.drawComplexControl(QStyle::CC_ToolButton, opt);
        // }

        if (adjSz != sz)
        {
            resize(sz);
            drawPixmapNextTime = true;
        }
        else
            mDrawPixmap = true; // transfer the pixmap to the widget now!
    }
    if (mDrawPixmap)
    {
        QPainter painter(this);
        painter.setTransform(transform);
        painter.drawPixmap(originPoint, mPixmap);

        drawPixmapNextTime = false;
    }

    mDrawPixmap = drawPixmapNextTime;
}

bool NdeTaskButton::hasDragAndDropHover() const
{
    return mDNDTimer->isActive();
}
//void NdeTaskButton::setButtonStatus(TASK_BUTTON_STATUS status)
//{
//    mButtonStatus = status;
//}
//NdeTaskButton::TASK_BUTTON_STATUS NdeTaskButton::buttonStatus() const
//{
//    return mButtonStatus;
//}
