/* BEGIN_COMMON_COPYRIGHT_HEADER
 * (c)LGPL2+
 *
 * Nde - a lightweight, Qt based, desktop toolset
 * https://lxqt.org
 *
 * Copyright: 2011 Razor team
 *            2014 Nde team
 * Authors:
 *   Alexander Sokoloff <sokoloff.a@gmail.com>
 *   Maciej Płaza <plaza.maciej@gmail.com>
 *   Kuzma Shapran <kuzma.shapran@gmail.com>
 *
 * This program or library is free software; you can redistribute it
 * and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301 USA
 *
 * END_COMMON_COPYRIGHT_HEADER */

#include "ndetaskgroup.h"
#include "ndetaskbar.h"
#include <QDebug>
#include <QMimeData>
#include <QDragLeaveEvent>
#include <QMenu>
#include <XdgIcon>
#include <KF5/KWindowSystem/KWindowSystem>
#include <functional>
#include "../panel/popupmenu.h"
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <QX11Info>
#include <QApplication>
#include <QScreen>
#include <QProcess>
#include <QTimer>
#include "../panel/thumbnailmanager.h"

/************************************************

 ************************************************/
NdeTaskGroup::NdeTaskGroup(const QString &groupName, WId window, bool autoRotation, NdeTaskBar *parent)
    : NdeTaskButton(window, parent, autoRotation, parent),
    mGroupName(groupName),
    mPopup(new NdeGroupPopup(this)),
	mSingleButton(true)
{
    init(parent);
}
NdeTaskGroup::NdeTaskGroup(const XdgDesktopFile &xdg,  bool autoRotation, NdeTaskBar * parent):
    NdeTaskButton(xdg, parent, autoRotation, parent),
    mGroupName(""),
    mPopup(new NdeGroupPopup(this)),
    mSingleButton(true)
{
    init(parent);
}
void NdeTaskGroup::init(NdeTaskBar *parent)
{
    Q_ASSERT(parent);

    setObjectName(mGroupName);
    setText(mGroupName);

    open_timer = new QTimer(this);
    open_timer->setSingleShot(true);
    connect(this, SIGNAL(clicked(bool)), this, SLOT(onClicked(bool)));
    connect(KWindowSystem::self(), SIGNAL(currentDesktopChanged(int)), this, SLOT(onDesktopChanged(int)));
    connect(KWindowSystem::self(), SIGNAL(activeWindowChanged(WId)), this, SLOT(onActiveWindowChanged(WId)));
    connect(parent, &NdeTaskBar::buttonRotationRefreshed, this, &NdeTaskGroup::setAutoRotation);
    connect(parent, &NdeTaskBar::refreshIconGeometry, this, &NdeTaskGroup::refreshIconsGeometry);
    connect(parent, &NdeTaskBar::buttonStyleRefreshed, this, &NdeTaskGroup::setToolButtonsStyle);
    connect(parent, &NdeTaskBar::showOnlySettingChanged, this, &NdeTaskGroup::refreshVisibility);
    connect(parent, &NdeTaskBar::popupShown, this, &NdeTaskGroup::groupPopupShown);
    connect(open_timer, SIGNAL(timeout()), this, SLOT(handleTime()));

    mAnimation = new QPropertyAnimation(mPopup, "geometry", mPopup);
    mAnimation->setDuration(300);
    mAnimation->setEasingCurve(QEasingCurve::OutQuart);
}
NdeTaskGroup::~NdeTaskGroup()
{
}

void NdeTaskGroup::realign()
{
    mPopup->realign();
}

QList<WId> NdeTaskGroup::buttonsWId() const
{
    return mButtonHash.keys();
}
/************************************************

 ************************************************/
void NdeTaskGroup::contextMenuEvent(QContextMenuEvent *event)
{
    setPopupVisible(false, true);
    mPreventPopup = true;
    if (mSingleButton)
    {
        NdeTaskButton::contextMenuEvent(event);
        return;
    }

    QMenu * menu = new PopupMenu(mPlugin->panel()->compositorOpen(), this);
    menu->setAttribute(Qt::WA_DeleteOnClose);
//    if(isFixed())
//    {
//        QAction *a = menu->addAction(tr("Unpin To TaskBar"));
//        connect(a, SIGNAL(triggered()), this, SLOT(unpinApp()));
//    }
//    if(buttonStatus() != BUTTON_FIXED)
//    {
//        QAction *a = menu->addAction(tr("Close group"));
//        connect(a, SIGNAL(triggered()), this, SLOT(closeGroup()));

//    }
    QAction *a = menu->addAction(tr("Close group"));
    connect(a, SIGNAL(triggered()), this, SLOT(closeGroup()));
    connect(menu, &QMenu::aboutToHide, [this] {
        mPreventPopup = false;
        update();
    });
    menu->setGeometry(plugin()->panel()->calculatePopupWindowPos(mapToGlobal(event->pos()), menu->sizeHint()));
    plugin()->willShowWindow(menu);
    menu->exec();
}

/************************************************

 ************************************************/
void NdeTaskGroup::closeGroup()
{
    for (NdeThumbnailButton *button : qAsConst(mButtonHash) )
        if (button->isOnDesktop(KWindowSystem::currentDesktop()))
            button->closeApplication();
}

/************************************************

 ************************************************/
NdeThumbnailButton * NdeTaskGroup::addWindow(WId id)
{
    if (mButtonHash.contains(id))
        return mButtonHash.value(id);

    NdeThumbnailButton *btn = new NdeThumbnailButton(id, parentTaskBar(), mPopup);
//    if (btn->isApplicationActive())
//    {
//        btn->setChecked(true);
//        setChecked(true);
//    }

    mButtonHash.insert(id, btn);
    mPopup->addButton(btn);
    connect(btn, SIGNAL(clicked()), this, SLOT(onChildButtonClicked()));
    refreshVisibility();

    updateThumbnail();
    return btn;
}

/************************************************

 ************************************************/
NdeThumbnailButton * NdeTaskGroup::checkedButton() const
{
    for (NdeThumbnailButton* button : qAsConst(mButtonHash))
    {
        if(button->isApplicationActive())
            return button;
    }
    return nullptr;
}

/************************************************

 ************************************************/
NdeThumbnailButton * NdeTaskGroup::getNextPrevChildButton(bool next, bool circular)
{
    NdeThumbnailButton *button = checkedButton();
    int idx = mPopup->indexOf(button);
    int inc = next ? 1 : -1;
    idx += inc;

    // if there is no cheked button, get the first one if next equals true
    // or the last one if not
    if (!button)
    {
        idx = -1;
        if (next)
        {
            for (int i = 0; i < mPopup->count() && idx == -1; i++)
                if (mPopup->itemAt(i)->widget()->isVisibleTo(mPopup))
                    idx = i;
        }
        else
        {
            for (int i = mPopup->count() - 1; i >= 0 && idx == -1; i--)
                if (mPopup->itemAt(i)->widget()->isVisibleTo(mPopup))
                    idx = i;
        }
    }

    if (circular)
        idx = (idx + mButtonHash.count()) % mButtonHash.count();
    else if (mPopup->count() <= idx || idx < 0)
        return nullptr;

    // return the next or the previous child
    QLayoutItem *item = mPopup->itemAt(idx);
    if (item)
    {
        button = qobject_cast<NdeThumbnailButton*>(item->widget());
        if (button->isVisibleTo(mPopup))
            return button;
    }

    return nullptr;
}

/************************************************

 ************************************************/
void NdeTaskGroup::onActiveWindowChanged(WId window)
{
    NdeThumbnailButton *button = mButtonHash.value(window, nullptr);
//    for (NdeThumbnailButton *btn : qAsConst(mButtonHash))
//        btn->setChecked(false);

    if (button)
    {
//        button->setChecked(true);
        if (button->hasUrgencyHint())
            button->setUrgencyHint(false);
    }
    setChecked(nullptr != button);
}

/************************************************

 ************************************************/
void NdeTaskGroup::onDesktopChanged(int number)
{
    refreshVisibility();
}

/************************************************

 ************************************************/
void NdeTaskGroup::onWindowRemoved(WId window)
{
    QMap<WId, QDateTime>::iterator it;
    it = m_freReduc.find(window);
    if (it != m_freReduc.end()){
        m_freReduc.erase(it);
    }

    if (mButtonHash.contains(window))
    {
        NdeThumbnailButton *button = mButtonHash.value(window);
        mButtonHash.remove(window);
        if(visibleButtonsCount() < 1)
            setPopupVisible(false, true);
        mPopup->removeWidget(button);
        //防止mPopup析构时button作为子窗口被重复析构
        button->setParent(nullptr);
        button->deleteLater();
        button = nullptr;

        if (mButtonHash.count())
            regroup();
        else
        {
            if (isVisible())
                Q_EMIT visibilityChanged(false);
//            if(buttonStatus() != BUTTON_FIXED)

//            {
//                hide();
//                Q_EMIT groupBecomeEmpty(groupName());
//            }
            hide();
            Q_EMIT groupBecomeEmpty(groupName());
        }
    }
}

/************************************************

 ************************************************/
void NdeTaskGroup::onChildButtonClicked()
{
    setPopupVisible(false, true);
}

/************************************************

 ************************************************/
void NdeTaskGroup::setToolButtonsStyle(Qt::ToolButtonStyle style)
{
    setToolButtonStyle(style);
}

/************************************************

 ************************************************/
int NdeTaskGroup::buttonsCount() const
{
    return mButtonHash.count();
}

/************************************************

 ************************************************/
int NdeTaskGroup::visibleButtonsCount() const
{
    int i = 0;
    for (NdeThumbnailButton *btn : qAsConst(mButtonHash))
        if (btn->isVisibleTo(mPopup))
            i++;
    return i;
}

/************************************************

 ************************************************/
void NdeTaskGroup::draggingTimerTimeout()
{
    if (mSingleButton)
        setPopupVisible(false);
}

/************************************************

 ************************************************/
void NdeTaskGroup::onClicked(bool)
{
    if (visibleButtonsCount() > 1)
    {
        setChecked(mButtonHash.contains(KWindowSystem::activeWindow()));
    }
}

/************************************************

 ************************************************/
void NdeTaskGroup::regroup()
{
    int cont = visibleButtonsCount();
    recalculateFrameIfVisible();
    if (cont == 0)
        hide();
    else if(cont == 1)
    {
        mSingleButton = true;
        // Get first visible button
        NdeThumbnailButton * button = NULL;
        for (NdeThumbnailButton *btn : qAsConst(mButtonHash))
        {
            if (btn->isVisibleTo(mPopup))
            {
                button = btn;
                break;
            }
        }

        if (button)
        {
            setText(button->title());
            setToolTip(button->toolTip());
            setWindowId(button->windowId());
        }
    }
    else
    {
        mSingleButton = false;
        QString t = QString("%1 - %2 windows").arg(mGroupName).arg(cont);
        setText(t);
        setToolTip(parentTaskBar()->isShowGroupOnHover() ? QString() : t);
    }
}

/************************************************

 ************************************************/
void NdeTaskGroup::recalculateFrameIfVisible()
{
    if (mPopup->isVisible())
    {
        if (plugin()->panel()->position() == INdePanel::PositionBottom)
            recalculateFramePosition();
    }
}

/************************************************

 ************************************************/
void NdeTaskGroup::setAutoRotation(bool value, INdePanel::Position position)
{
    for (NdeThumbnailButton *button : qAsConst(mButtonHash))
        button->setAutoRotation(false, position);
    mPopup->changeLayout();
    NdeTaskButton::setAutoRotation(value, position);
}

/************************************************

 ************************************************/
void NdeTaskGroup::refreshVisibility()
{
    bool will = false;
    NdeTaskBar const * taskbar = parentTaskBar();
    const int showDesktop = taskbar->showDesktopNum();
    for(NdeThumbnailButton * btn : qAsConst(mButtonHash))
    {
//        bool visible = taskbar->isShowOnlyOneDesktopTasks() ? btn->isOnDesktop(0 == showDesktop ? KWindowSystem::currentDesktop() : showDesktop) : true;
//        visible &= taskbar->isShowOnlyCurrentScreenTasks() ? btn->isOnCurrentScreen() : true;
        //这里强制只显示当前屏幕和当前桌面中的应用
        bool visible = btn->isOnDesktop(0 == showDesktop ? KWindowSystem::currentDesktop() : showDesktop);
        //这里注销掉，在多屏模式下在副屏的应用也应该显示出来
//        visible &= btn->isOnCurrentScreen();

//        visible &= taskbar->isShowOnlyMinimizedTasks() ? btn->isMinimized() : true;
        btn->setVisible(visible);
        will |= visible;
    }
//    if(buttonStatus() == BUTTON_FIXED)
//        will = true;
    bool is = isVisible();
    setVisible(will);
    regroup();

    if (is != will)
        Q_EMIT visibilityChanged(will);
}

/************************************************

 ************************************************/
QMimeData *NdeTaskGroup::mimeData()
{
    QMimeData *mimedata = new QMimeData;
    QByteArray byteArray;
    QDataStream stream(&byteArray, QIODevice::WriteOnly);
    stream << groupName();
    mimedata->setData(mimeDataFormat(), byteArray);
    return mimedata;
}

/************************************************

 ************************************************/
void NdeTaskGroup::setPopupVisible(bool visible, bool fast)
{

    int count = visibleButtonsCount();
    if (visible && !mPreventPopup && count >= 1)
    {
        if (!mPopup->isVisible())
        {
            recalculateFramePosition();
        }
        updateThumbnail();
        plugin()->willShowWindow(mPopup);
        mPopup->show();
        Q_EMIT popupShown(this);


        //        mAnimation->start();
    }
    else
    {
        //        mAnimation->stop();
        mPopup->hide(fast);
    }
}

NdeThumbnailButton * NdeTaskGroup::findNdeThumbailButton(WId id)
{
    auto it = mButtonHash.find(id);
    if (mButtonHash.end() != it)
    {
        return it.value();
    }

    return nullptr;
}

/************************************************

 ************************************************/
void NdeTaskGroup::refreshIconsGeometry()
{
    updateThumbnail();

    QRect rect = geometry();
    rect.moveTo(mapToGlobal(QPoint(0, 0)));

    if (mSingleButton)
    {
        refreshIconGeometry(rect);
        return;
    }
}

/************************************************

 ************************************************/
int NdeTaskGroup::recalculateFrameHeight() const
{
    int cont = visibleButtonsCount();
    if(plugin()->panel()->isHorizontal() )
        return PopupMenuHeight;
    else
    {
        return cont * PopupMenuHeight + (cont + 1) * mPopup->spacing();
    }

}

/************************************************

 ************************************************/
int NdeTaskGroup::recalculateFrameWidth() const
{
    int cont = visibleButtonsCount();
    if(plugin()->panel()->isHorizontal() )
        return cont * PopupMenuHeight + (cont + 1) * mPopup->spacing();
    else
    {
        return PopupMenuWidth;
    }
}

QSize NdeTaskGroup::recalculateFrameSize()
{
    int height = recalculateFrameHeight();
    mPopup->setMaximumHeight(height);
    mPopup->setMinimumHeight(0);

    int width = recalculateFrameWidth();
    mPopup->setMaximumWidth(width);
    mPopup->setMinimumWidth(0);
    QRect rc = qApp->primaryScreen()->availableGeometry();
    switch (plugin()->panel()->position()) {
    case INdePanel::PositionTop:
    case INdePanel::PositionBottom:
    {
        if(width > rc.width())
        {
            width = rc.width();
            int cont = visibleButtonsCount();
            height = rc.width() / cont * PopupMenuHeight / PopupMenuWidth;
        }
    }
        break;
    case INdePanel::PositionLeft:
    case INdePanel::PositionRight:
    {
        if(height > rc.height())
        {
            height = rc.height();
            int cont = visibleButtonsCount();
            width = rc.height() / cont * PopupMenuWidth / PopupMenuHeight;
        }
        break;
    }
    }
    QSize newSize(width, height);
    mPopup->resize(newSize);

    return newSize;
}
QPoint NdeTaskGroup::recalculateFramePosition()
{
    QSize newSize = recalculateFrameSize();
    int x_start = this->rect().center().x();
    int y_start = this->rect().center().y();
    int x_end = 0;
    int y_end = 0;
    int w_start = 0;
    int h_start = 0;
    switch (plugin()->panel()->position())
    {
    case INdePanel::PositionTop:
        x_start -= newSize.width()/2;
        y_start += height()/2;
        x_end = x_start;
        y_end += y_start + 4;
        w_start = newSize.width();
        break;
    case INdePanel::PositionBottom:
        x_start -= newSize.width()/2;
        y_start -= height()/2;
        x_end = x_start;
        y_end += y_start - newSize.height() - 4;
        w_start = newSize.width();
        break;
    case INdePanel::PositionLeft:
        x_start += width()/2;
        y_start -= newSize.height()/2;
        x_end += x_start + 4;
        y_end = y_start;
        h_start = newSize.height();
        break;
    case INdePanel::PositionRight:
        x_start -= width()/2;
        y_start -= newSize.height()/2;
        x_end += x_start - newSize.width() - 3;
        y_end = y_start;
        h_start = newSize.height();
        break;
    }
    QPoint endPos = mapToGlobal(QPoint(x_end, y_end));

    QPoint startPos = mapToGlobal(QPoint(x_start, y_start));
    int w_end = newSize.width();
    int h_end = newSize.height();
    x_start = startPos.x() <= 0 ? 0 : startPos.x();


    x_end = endPos.x() <= 0 ? 0 : endPos.x();
    y_end = endPos.y() <= 0 ? 0 : endPos.y();

    //    mAnimation->setStartValue(QRect(x_start,
    //                                    startPos.y(),
    //                                    w_start, h_start));

    //    mAnimation->setEndValue(QRect(x_end,
    //                                  endPos.y(),
    //                                  w_end, h_end));
    mPopup->move(QPoint(x_end, y_end));
    return endPos;
}

/************************************************

 ************************************************/
void NdeTaskGroup::leaveEvent(QEvent *event)
{
    open_timer->stop();
    setPopupVisible(false);
    NdeTaskButton::leaveEvent(event);
}

/************************************************

 ************************************************/
void NdeTaskGroup::enterEvent(QEvent *event)
{
    NdeTaskButton::enterEvent(event);
    open_timer->start(200);
}

void NdeTaskGroup::handleTime(){
    if (sDraggging)
        return;
    if (parentTaskBar()->isShowGroupOnHover()){
        setPopupVisible(true);
    }
}

/************************************************

 ************************************************/
void NdeTaskGroup::dragEnterEvent(QDragEnterEvent *event)
{
    // only show the popup if we aren't dragging a taskgroup
    if (!event->mimeData()->hasFormat(mimeDataFormat()))
    {
        setPopupVisible(true);
    }
    NdeTaskButton::dragEnterEvent(event);
}

/************************************************

 ************************************************/
void NdeTaskGroup::dragLeaveEvent(QDragLeaveEvent *event)
{
    // if draggind something into the taskgroup or the taskgroups' popup,
    // do not close the popup
    if (!sDraggging)
        setPopupVisible(false);
    NdeTaskButton::dragLeaveEvent(event);
}

void NdeTaskGroup::mouseMoveEvent(QMouseEvent* event)
{
    // if dragging the taskgroup, do not show the popup
    setPopupVisible(false, true);
    NdeTaskButton::mouseMoveEvent(event);
}

void NdeTaskGroup::mousePressEvent(QMouseEvent *event)
{
    if(mPopup->isVisible())
    {
        setPopupVisible(false, true) ;
    }
    NdeTaskButton::mousePressEvent(event);
}
void NdeTaskGroup::mouseReleaseEvent(QMouseEvent *event)
{
    int cont = visibleButtonsCount();
    if(isMoved)
    {
        isMoved = false;
        QToolButton::mouseReleaseEvent(event);
        return;
    }
    if(cont <= 1)
    {
        NdeTaskButton::mouseReleaseEvent(event);
    }
    else
    {
//        if(mPopup->isVisible())
//            setPopupVisible(false, true) ;
//        else
        setPopupVisible(true);
        QToolButton::mouseReleaseEvent(event);
    }
}
/************************************************

 ************************************************/
bool NdeTaskGroup::onWindowChanged(WId window, NET::Properties prop, NET::Properties2 prop2)
{ // returns true if the class is preserved
    bool needsRefreshVisibility{false};
    QVector<QWidget *> buttons;
    if (mButtonHash.contains(window))
        buttons.append(mButtonHash.value(window));

    // If group is based on that window properties must be changed also on button group
    if (window == windowId())
        buttons.append(this);

    if (!buttons.isEmpty())
    {
        // if class is changed the window won't belong to our group any more
        if (parentTaskBar()->isGroupingEnabled() && prop2.testFlag(NET::WM2WindowClass))
        {
            KWindowInfo info(window, nullptr, NET::WM2WindowClass);
            if (info.windowClassClass() != mGroupName)
            {
                onWindowRemoved(window);
                return false;
            }
        }
        // window changed virtual desktop
        if (prop.testFlag(NET::WMDesktop) || prop.testFlag(NET::WMGeometry))
        {
//            if (parentTaskBar()->isShowOnlyOneDesktopTasks()
//                    || parentTaskBar()->isShowOnlyCurrentScreenTasks())
//            {
//                needsRefreshVisibility = true;
//            }
            needsRefreshVisibility = true;
        }

//        if (prop.testFlag(NET::WMVisibleName) || prop.testFlag(NET::WMName))
//            std::for_each(buttons.begin(), buttons.end(), std::mem_fn(&NdeTaskButton::updateText));

        // XXX: we are setting window icon geometry -> don't need to handle NET::WMIconGeometry
        // Icon of the button can be based on windowClass
//        if (prop.testFlag(NET::WMIcon) || prop2.testFlag(NET::WM2WindowClass))
//            std::for_each(buttons.begin(), buttons.end(), std::mem_fn(&NdeTaskButton::updateIcon));

        if (prop.testFlag(NET::WMState))
        {
            KWindowInfo info{window, NET::WMState};
            if (info.hasState(NET::SkipTaskbar))
                onWindowRemoved(window);
//            std::for_each(buttons.begin(), buttons.end(), std::bind(&NdeTaskButton::setUrgencyHint, std::placeholders::_1, info.hasState(NET::DemandsAttention)));

//            if (parentTaskBar()->isShowOnlyMinimizedTasks())
//            {
//                needsRefreshVisibility = true;
//            }
            needsRefreshVisibility = true;
        }
    }

    updateThumbnail();
    if (needsRefreshVisibility) {
        refreshVisibility();
    }
    else {
        //更新taskbtn的名称
        int cont = visibleButtonsCount();
        if(cont == 1)
        {
            // Get first visible button
            NdeThumbnailButton * button = NULL;
            for (NdeThumbnailButton *btn : qAsConst(mButtonHash))
            {
                if (btn->isVisibleTo(mPopup))
                {
                    button = btn;
                    break;
                }
            }

            if (button)
            {
                setText(button->title());
                setToolTip(button->toolTip());
                setWindowId(button->windowId());
            }
        }
        else if(cont > 1)
        {
            QString t = QString("%1 - %2 windows").arg(mGroupName).arg(cont);
            setText(t);
            setToolTip(parentTaskBar()->isShowGroupOnHover() ? QString() : t);
        }
    }

    return true;
}

bool NdeTaskGroup::onWindowDamaged(WId id, const uint8_t responseType)
{
    auto it = mButtonHash.find(id);
    if(it != mButtonHash.end() && it.value() != nullptr)
    {
        void* image = ThumbnailManager::getIns()->getXimage(it.value()->windowId());
        it.value()->addThumbnail(image);
        it.value()->onWindowDamaged(responseType);
    }
    return false;
}

/************************************************

 ************************************************/
void NdeTaskGroup::groupPopupShown(NdeTaskGroup * const sender)
{
    //close all popups (should they be visible because of close delay)
    if (this != sender && isVisible())
            setPopupVisible(false, true/*fast*/);
}

void NdeTaskGroup::updateThumbnail()
{
    for(NdeThumbnailButton * btn : qAsConst(mButtonHash)){
        //mPopup->addButton(btn);
        btn->updateTitle();

        QDateTime now = QDateTime::currentDateTime();
        if (m_freReduc.contains(btn->windowId()))
        {
            QDateTime last = m_freReduc.value(btn->windowId());
            if (last.msecsTo(now) < 200)
            {
                return ;
            }
        }
        m_freReduc.insert(btn->windowId(), now);

//        QPixmap thumbnail = ThumbnailManager::getIns()->getThumbnail(btn->windowId());
//        thumbnail.scaled(PopupMenuWidth,PopupMenuHeight,Qt::KeepAspectRatio,Qt::SmoothTransformation);
//        void* image = ThumbnailManager::getIns()->getXimage(btn->windowId());
//        btn->addThumbnail(image);
    }
}
